/*
 * coff-pic30.c -- BFD back-end for Microchip PIC30 COFF binaries.
 * Copyright (C) 2000 Free Software Foundation, Inc.
 * Contributed by Microchip Corporation.
 * Written by Tracy A. Kuhrt
 * Modified by John Elliott (john.elliott@microchip.com)
 *
 * This file is part of BFD, the Binary File Descriptor library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "bfd.h"
#include "sysdep.h"
#include "libbfd.h"
#include "coff/pic30.h"
#include "coff/internal.h"
#include "libcoff.h"
#include <coff-pic30.h>

/* Data Structures for the Handle Jump Table */
bfd *handle_bfd;
unsigned char *handle_data;
struct handle_hash_table *handles;

/* Data Structures for the Data Init Template */
bfd *init_bfd;
unsigned char *init_data;

/* Data Structures for the Default Interrupt Handler */
bfd *isr_bfd;
unsigned char *isr_data;

/* Variables set by pic30 linker options */
boolean pic30_debug = 0;
boolean pic30_data_init = true;
boolean pic30_has_data_init_option = 0;
boolean pic30_handles = true;
boolean pic30_has_handles_option = 0;
boolean pic30_pack_data = true;
boolean pic30_has_pack_data_option = 0;
boolean pic30_has_stack_option = 0;
unsigned int pic30_stack_size = 16;
boolean pic30_has_heap_option = 0;
boolean pic30_heap_required = 0;
unsigned int pic30_heap_size = 0;
boolean pic30_smart_io = true;
boolean pic30_has_smart_io_option = 0;

/* Other state variables */
boolean pic30_has_user_startup = 0;
boolean pic30_has_user_default_interrupt = 0;

extern void pic30_rtype2howto
  PARAMS ((arelent *, struct internal_reloc *));

extern bfd_reloc_status_type pic30_bfd_reloc_range_check
  PARAMS ((reloc_howto_type *, bfd_vma, char **));

static void bfd_pic30_report_sections
  PARAMS ((bfd *, asection *, PTR));

static boolean pic30_final_link
  PARAMS ( (bfd *, struct bfd_link_info *));

static boolean pic30_set_section_contents
  PARAMS ((bfd *, sec_ptr, PTR, file_ptr, bfd_size_type));

static reloc_howto_type * pic30_bfd_reloc_type_lookup
  PARAMS ((bfd *, bfd_reloc_code_real_type));

static bfd_reloc_status_type pic30_coff_is_even_address
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_reloc_status_type pic30_coff_is_file_reg_address
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_reloc_status_type pic30_coff_perform_program_address
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_reloc_status_type pic30_coff_perform_pc_relative
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_vma pic30_coff_extract_bytes
   PARAMS ((bfd *, PTR, int, bfd_size_type, unsigned char, unsigned char));

static void pic30_coff_insert_bytes PARAMS ((bfd *, PTR, int,
   bfd_size_type, bfd_vma, unsigned char, unsigned char));

static bfd_vma pic30_coff_extract_bytes_data_mem
   PARAMS ((bfd *, PTR, int, bfd_size_type));

static void pic30_coff_insert_bytes_data_mem
   PARAMS ((bfd *, PTR, int, bfd_size_type, bfd_vma));

static bfd_reloc_status_type pic30_coff_perform_data_directive
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_reloc_status_type pic30_coff_perform_p_data_directive
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_vma bfd_pic30_handle
     PARAMS ((bfd *, bfd_vma, asymbol *));

static bfd_reloc_status_type pic30_coff_perform_operators
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_reloc_status_type pic30_coff_perform_p_operators
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

static bfd_reloc_status_type pic30_coff_perform_generic
   PARAMS ((bfd *, arelent *, asymbol *, PTR, asection *, bfd *, char **));

#define RELOC_SPECIAL_FN_EVEN_ADDRESS pic30_coff_is_even_address
#define RELOC_SPECIAL_FN_FILE_REG_ADDRESS pic30_coff_is_file_reg_address
#define RELOC_SPECIAL_FN_PGM_ADDR pic30_coff_perform_program_address
#define RELOC_SPECIAL_FN_PCREL pic30_coff_perform_pc_relative
#define RELOC_SPECIAL_FN_DATA_DIRECTIVE pic30_coff_perform_data_directive
#define RELOC_SPECIAL_FN_P_DATA_DIRECTIVE pic30_coff_perform_p_data_directive
#define RELOC_SPECIAL_FN_OPERATORS pic30_coff_perform_operators
#define RELOC_SPECIAL_FN_P_OPERATORS pic30_coff_perform_p_operators
#define RELOC_SPECIAL_FN_GENERIC pic30_coff_perform_generic

#define COFF_DEFAULT_SECTION_ALIGNMENT_POWER (1)
#define	COFF_LONG_FILENAMES 1
#define	COFF_LONG_SECTION_NAMES 1

/* PIC30 COFF encodes the section alignment in the section header flags */
#define COFF_ALIGN_IN_SECTION_HEADER 1
#define COFF_ALIGN_IN_S_FLAGS 1
#define COFF_ENCODE_ALIGNMENT(S,X) ((S).s_flags |= (((unsigned)(X)&0xF)<<28))
#define COFF_DECODE_ALIGNMENT(X) (((X)>>28)&0xF)

#define DATA_UPPER_FILL_BYTE 0xFF


reloc_howto_type pic30_coff_howto_table[] =
{
/* type 
 * rightshift - Value the final relocation is shifted right by
 * size (0 = byte, 1 = short, 2 = long)
 * bitsize - Number of bits in the item to be relocated.
 * pc_relative - Boolean specifying if the relocation is relative to the
 *               location
 * bitpos - Relocated value is left shifted by this much
 * complain_on_overflow - Type of overflow error that should be checked for
 * special_function - This function is called when not NULL.
 * name - textual name
 * partial_inplace - When performing partial link, do you modify the
 *                   relocation instead of the data.
 * src_mask - Which parts of the read in data are to be used in the
 *            relocation sum.
 * dst_mask - Selects which parts of the instruction are replaced into
 *            the instruction.
 * pcrel_offset - Signals whether the value of the pc is left in the slot of
 *                the instruction.
 */
   HOWTO(R_PIC30_FILE_REG_BYTE, 0, 1, 13, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_FILE_REG_ADDRESS, "FILE REG - BYTE",
         true, 0x01fff, 0x001fff, false),
   HOWTO(R_PIC30_FILE_REG, 0, 1, 13, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_FILE_REG_ADDRESS, "FILE REG",
         true, 0x01fff, 0x001fff, false),
   HOWTO(R_PIC30_FILE_REG_WORD, 0, 1, 16, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_EVEN_ADDRESS, "FILE REG - WORD",
         true, 0x0ffff, 0x00ffff, false),
   HOWTO(R_PIC30_FILE_REG_WORD_WITH_DST, 1, 2, 15, false, 4,
         complain_overflow_unsigned,
         RELOC_SPECIAL_FN_EVEN_ADDRESS, "FILE REG - WORD with DST.",
         true, 0x07fff0, 0x07fff0, false),
   HOWTO(R_PIC30_UNSIGNED_4, 0, 0, 4, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "UNSIGNED 4",
         true, 0x000f, 0x00000f, false),
   HOWTO(R_PIC30_UNSIGNED_5, 0, 0, 5, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "UNSIGNED 5",
         true, 0x001f, 0x00001f, false),
   HOWTO(R_PIC30_PCREL_DO, 1, 1, 16, true, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_PCREL, "PC RELATIVE DO",
         true, 0xffff, 0x0000ffff, true),
   HOWTO(R_PIC30_DO_ABSOLUTE, 1, 1, 16, true, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_PCREL, "DO ABSOLUTE",
         true, 0xffff, 0x0000ffff, true),
   HOWTO(R_PIC30_SIGNED_10_BYTE, 0, 1, 10, false, 4, complain_overflow_dont,
         RELOC_SPECIAL_FN_GENERIC, "SIGNED 10 (byte)",
         true, 0x000ff0, 0x000ff0, false),
   HOWTO(R_PIC30_UNSIGNED_10, 0, 1, 10, false, 4, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "UNSIGNED 10",
         true, 0x003ff0, 0x003ff0, false),
   HOWTO(R_PIC30_UNSIGNED_14, 0, 1, 14, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "UNSIGNED 14",
         true, 0x003fff, 0x003fff, false),
   HOWTO(R_PIC30_FRAME_SIZE, 0, 1, 14, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_EVEN_ADDRESS, "FRAME SIZE",
         true, 0x003fff, 0x003fff, false),
   HOWTO(R_PIC30_WORD, 0, 2, 16, false, 4, complain_overflow_dont,
         RELOC_SPECIAL_FN_GENERIC, "WORD",
         true, 0xffff0, 0x0ffff0, false),
   HOWTO(R_PIC30_WORD_TBLOFFSET, 0, 2, 16, false, 4, complain_overflow_dont,
         RELOC_SPECIAL_FN_OPERATORS, "WORD - TBLOFFSET",
         true, 0xffff0, 0x0ffff0, false),
   HOWTO(R_PIC30_WORD_PSVOFFSET, 0, 2, 16, false, 4, complain_overflow_dont,
         RELOC_SPECIAL_FN_OPERATORS, "WORD - PSVOFFSET",
         true, 0xffff0, 0x0ffff0, false),
   HOWTO(R_PIC30_WORD_TBLPAGE, 16, 2, 16, false, 4, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_OPERATORS, "WORD - TBLPAGE",
         true, 0xffff0, 0x0ffff0, false),
   HOWTO(R_PIC30_WORD_PSVPAGE, 15, 2, 16, false, 4, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_OPERATORS, "WORD - PSVPAGE",
         true, 0xffff0, 0x0ffff0, false),
   HOWTO(R_PIC30_WORD_HANDLE, 0, 2, 16, false, 4, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_OPERATORS, "WORD - HANDLE",
         true, 0xffff0, 0x0ffff0, false),
   HOWTO(R_PIC30_PCREL_BRANCH, 1, 1, 16, true, 0, complain_overflow_signed,
         RELOC_SPECIAL_FN_PCREL, "PC RELATIVE BRANCH",
         true, 0xffff, 0x00ffff, true),
   HOWTO(R_PIC30_BRANCH_ABSOLUTE, 1, 1, 16, true, 0, complain_overflow_signed,
         RELOC_SPECIAL_FN_PCREL, "BRANCH ABSOLUTE",
         true, 0xffff, 0x00ffff, true),
   HOWTO(R_PIC30_PGM_ADDR_LSB, 0, 1, 16, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_PGM_ADDR, "PROGRAM ADDRESS LSB",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_PGM_ADDR_MSB, 16, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_PGM_ADDR, "PROGRAM ADDRESS MSB",
         true, 0xffff, 0xffff, false),
   HOWTO(R_PIC30_TBLOFFSET, 0, 1, 16, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_OPERATORS, "TBLOFFSET",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_PSVOFFSET, 0, 1, 16, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_OPERATORS, "PSVOFFSET",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_TBLPAGE, 16, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_OPERATORS, "TBLPAGE",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_PSVPAGE, 15, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_OPERATORS, "PSVPAGE",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_HANDLE, 0, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_OPERATORS, "HANDLE (pointer to function)",
         true, 0x0000ffff, 0x0000ffff, false),
   /* This relocation can only be created via a .long PADDR(foo) statement */
   HOWTO(R_PIC30_PADDR, 0, 2, 32, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_OPERATORS, "PADDR",
         true, 0xffffffff, 0x00ffffff, false),
   HOWTO(R_PIC30_P_TBLOFFSET, 0, 1, 16, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_P_OPERATORS, "P-TBLOFFSET",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_P_PSVOFFSET, 0, 1, 16, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_P_OPERATORS, "P-PSVOFFSET",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_P_TBLPAGE, 16, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_P_OPERATORS, "P-TBLPAGE",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_P_PSVPAGE, 15, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_P_OPERATORS, "P-PSVPAGE",
         true, 0x0000ffff, 0x0000ffff, false),
   HOWTO(R_PIC30_P_HANDLE, 0, 1, 16, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_P_OPERATORS, "P-HANDLE",
         true, 0x0000ffff, 0x0000ffff, false),
   /* This relocation can only be created via a .pword PADDR(foo) statement */
   HOWTO(R_PIC30_P_PADDR, 0, 2, 24, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_P_OPERATORS, "P-PADDR",
         true, 0x00ffffff, 0x00ffffff, false),
   HOWTO(R_PIC30_BIT_SELECT_3, 0, 1, 3, false, 13, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "BIT SELECT 3",
         true, 0x00e000, 0x00e000, false),
   HOWTO(R_PIC30_BIT_SELECT_4_BYTE, 0, 1, 3, false, 12,
         complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "BIT SELECT 4 (byte)",
         true, 0x007000, 0x007000, false),
   HOWTO(R_PIC30_BIT_SELECT_4, 0, 1, 4, false, 12, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "BIT SELECT 4",
         true, 0x00f000, 0x00f000, false),
   HOWTO(R_PIC30_DSP_6, 0, 0, 5, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_GENERIC, "DSP SIGNED 6",
         true, 0x003f, 0x00003f, false),
   HOWTO(R_PIC30_DSP_PRESHIFT, 0, 1, 4, false, 7, complain_overflow_signed,
         RELOC_SPECIAL_FN_GENERIC, "DSP PRESHIFT",
         true, 0x000780, 0x000780, false),
   HOWTO(R_PIC30_PWRSAV_MODE, 0, 0, 1, false, 0, complain_overflow_unsigned,
         RELOC_SPECIAL_FN_GENERIC, "POWER SAVE MODE",
         true, 0x000001, 0x000001, false),
   HOWTO(R_PIC30_PBYTE, 0, 0, 8, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_P_DATA_DIRECTIVE, ".pbyte",
         true, 0x000000ff, 0x000000ff, false),
   HOWTO(R_PIC30_PWORD, 0, 2, 24, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_P_DATA_DIRECTIVE, ".pword",
         true, 0x00ffffff, 0x00ffffff, false),
   HOWTO(R_RELBYTE, 0,  0,   8,  false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_DATA_DIRECTIVE, ".byte",
         true, 0x000000ff,0x000000ff, false),
   HOWTO(R_RELWORD, 0, 1, 16, false, 0, complain_overflow_dont,
         RELOC_SPECIAL_FN_DATA_DIRECTIVE, ".word",
         true, 0x0000ffff,0x0000ffff, false),
   HOWTO(R_RELLONG, 0, 2, 32, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_DATA_DIRECTIVE, ".long",
         true, 0xffffffff,0xffffffff, false),
#ifdef BFD64
   HOWTO(R_PIC30_QUAD, 0, 4, 64, false, 0, complain_overflow_bitfield,
         RELOC_SPECIAL_FN_GENERIC, ".quad",
         true, 0xffffffffffffffff,0xffffffffffffffff, false)
#endif
};

/******************************************************************************/

int
pic30_adjustable_against_section (type)
   unsigned short type;
{
   int rc = 1;

   switch (type)
   {
      case R_PIC30_PCREL_BRANCH:
      case R_PIC30_PCREL_DO:
      case R_PIC30_BRANCH_ABSOLUTE:
      case R_PIC30_DO_ABSOLUTE:
      case R_PIC30_PGM_ADDR_MSB:
      case R_PIC30_PGM_ADDR_LSB:
      case R_PIC30_TBLOFFSET:
      case R_PIC30_PSVOFFSET:
      case R_PIC30_TBLPAGE:
      case R_PIC30_PSVPAGE:
      case R_PIC30_HANDLE:
      case R_PIC30_PADDR:
      case R_PIC30_P_TBLOFFSET:
      case R_PIC30_P_PSVOFFSET:
      case R_PIC30_P_TBLPAGE:
      case R_PIC30_P_PSVPAGE:
      case R_PIC30_P_HANDLE:
      case R_PIC30_P_PADDR:
      case R_PIC30_WORD_TBLOFFSET:
      case R_PIC30_WORD_PSVOFFSET:
      case R_PIC30_WORD_PSVPAGE:
      case R_PIC30_WORD_TBLPAGE:
      case R_PIC30_WORD_HANDLE:
      case R_PIC30_WORD:
      case R_PIC30_FILE_REG_WORD:
      case R_PIC30_FILE_REG_WORD_WITH_DST:
      case R_PIC30_PBYTE:
      case R_PIC30_PWORD:
      case R_RELBYTE:
      case R_RELWORD:
      case R_RELLONG:
         rc = 0;
         break;

      default:
         break;
   } /* switch (type) */

   return rc;
} /* int pic30_adjustable_against_section(unsigned short) */

/******************************************************************************/


/* Code to turn an r_type into a howto ptr, uses the above howto table */

void pic30_rtype2howto (internal, dst)
   arelent * internal;
   struct internal_reloc * dst;
{
   switch (dst->r_type)
   {
      case R_PIC30_FILE_REG_BYTE:
         internal->howto = pic30_coff_howto_table + 0;
         break;

      case R_PIC30_FILE_REG:
         internal->howto = pic30_coff_howto_table + 1;
         break;

      case R_PIC30_FILE_REG_WORD:
         internal->howto = pic30_coff_howto_table + 2;
         break;

      case R_PIC30_FILE_REG_WORD_WITH_DST:
         internal->howto = pic30_coff_howto_table + 3;
         break;

      case R_PIC30_UNSIGNED_4:
         internal->howto = pic30_coff_howto_table + 4;
         break;

      case R_PIC30_UNSIGNED_5:
         internal->howto = pic30_coff_howto_table + 5;
         break;

      case R_PIC30_PCREL_DO:
         internal->howto = pic30_coff_howto_table + 6;
         break;

      case R_PIC30_DO_ABSOLUTE:
         internal->howto = pic30_coff_howto_table + 7;
         break;

      case R_PIC30_SIGNED_10_BYTE:
         internal->howto = pic30_coff_howto_table + 8;
         break;

      case R_PIC30_UNSIGNED_10:
         internal->howto = pic30_coff_howto_table + 9;
         break;

      case R_PIC30_UNSIGNED_14:
         internal->howto = pic30_coff_howto_table + 10;
         break;

      case R_PIC30_FRAME_SIZE:
         internal->howto = pic30_coff_howto_table + 11;
         break;

      case R_PIC30_WORD:
         internal->howto = pic30_coff_howto_table + 12;
         break;

      case R_PIC30_WORD_TBLOFFSET:
         internal->howto = pic30_coff_howto_table + 13;
         break;

      case R_PIC30_WORD_PSVOFFSET:
         internal->howto = pic30_coff_howto_table + 14;
         break;

      case R_PIC30_WORD_TBLPAGE:
         internal->howto = pic30_coff_howto_table + 15;
         break;

      case R_PIC30_WORD_PSVPAGE:
         internal->howto = pic30_coff_howto_table + 16;
         break;

      case R_PIC30_WORD_HANDLE:
         internal->howto = pic30_coff_howto_table + 17;
         break;

      case R_PIC30_PCREL_BRANCH:
         internal->howto = pic30_coff_howto_table + 18;
         break;

      case R_PIC30_BRANCH_ABSOLUTE:
         internal->howto = pic30_coff_howto_table + 19;
         break;

      case R_PIC30_PGM_ADDR_LSB:
         internal->howto = pic30_coff_howto_table + 20;
         break;

      case R_PIC30_PGM_ADDR_MSB:
         internal->howto = pic30_coff_howto_table + 21;
         break;

      case R_PIC30_TBLOFFSET:
         internal->howto = pic30_coff_howto_table + 22;
         break;

      case R_PIC30_PSVOFFSET:
         internal->howto = pic30_coff_howto_table + 23;
         break;

      case R_PIC30_TBLPAGE:
         internal->howto = pic30_coff_howto_table + 24;
         break;

      case R_PIC30_PSVPAGE:
         internal->howto = pic30_coff_howto_table + 25;
         break;

      case R_PIC30_HANDLE:
         internal->howto = pic30_coff_howto_table + 26;
         break;

      case R_PIC30_PADDR:
         internal->howto = pic30_coff_howto_table + 27;
         break;

      case R_PIC30_P_TBLOFFSET:
         internal->howto = pic30_coff_howto_table + 28;
         break;

      case R_PIC30_P_PSVOFFSET:
         internal->howto = pic30_coff_howto_table + 29;
         break;

      case R_PIC30_P_TBLPAGE:
         internal->howto = pic30_coff_howto_table + 30;
         break;

      case R_PIC30_P_PSVPAGE:
         internal->howto = pic30_coff_howto_table + 31;
         break;

      case R_PIC30_P_HANDLE:
         internal->howto = pic30_coff_howto_table + 32;
         break;

      case R_PIC30_P_PADDR:
         internal->howto = pic30_coff_howto_table + 33;
         break;

      case R_PIC30_BIT_SELECT_3:
         internal->howto = pic30_coff_howto_table + 34;
         break;

      case R_PIC30_BIT_SELECT_4_BYTE:
         internal->howto = pic30_coff_howto_table + 35;
         break;

      case R_PIC30_BIT_SELECT_4:
         internal->howto = pic30_coff_howto_table + 36;
         break;

      case R_PIC30_DSP_6:
         internal->howto = pic30_coff_howto_table + 37;
         break;

      case R_PIC30_DSP_PRESHIFT:
         internal->howto = pic30_coff_howto_table + 38;
         break;

      case R_PIC30_PWRSAV_MODE:
         internal->howto = pic30_coff_howto_table + 39;
         break;

      case R_PIC30_PBYTE:
         internal->howto = pic30_coff_howto_table + 40;
         break;

      case R_PIC30_PWORD:
         internal->howto = pic30_coff_howto_table + 41;
         break;

      case R_RELBYTE:
         internal->howto = pic30_coff_howto_table + 42;
         break;

      case R_RELWORD:
         internal->howto = pic30_coff_howto_table + 43;
         break;

      case R_RELLONG:
         internal->howto = pic30_coff_howto_table + 44;
         break;

#ifdef BFD64
      case R_PIC30_QUAD:
         internal->howto = pic30_coff_howto_table + 45;
         break;
#endif

      default:
         abort ();
         break;
   } /* switch (dst->r_type) */
} /* void pic30_rtype2howto(...) */

#define RTYPE2HOWTO(internal, relocentry) pic30_rtype2howto(internal,relocentry)

/* Turn a howto into a reloc number */
#define SELECT_RELOC(x,howto) x.r_type = howto->type

static reloc_howto_type * pic30_bfd_reloc_type_lookup (abfd, code)
   bfd * abfd ATTRIBUTE_UNUSED;
   bfd_reloc_code_real_type code;
{
   switch (code)
   {
      case BFD_RELOC_PIC30_FILE_REG_BYTE:
         return pic30_coff_howto_table + 0;
         break;

      case BFD_RELOC_PIC30_FILE_REG:
         return pic30_coff_howto_table + 1;
         break;

      case BFD_RELOC_PIC30_FILE_REG_WORD:
         return pic30_coff_howto_table + 2;
         break;

      case BFD_RELOC_PIC30_FILE_REG_WORD_WITH_DST:
         return pic30_coff_howto_table + 3;
         break;

      case BFD_RELOC_PIC30_UNSIGNED_4:
         return pic30_coff_howto_table + 4;
         break;

      case BFD_RELOC_PIC30_UNSIGNED_5:
         return pic30_coff_howto_table + 5;
         break;

      case BFD_RELOC_PIC30_PCREL_DO:
         return pic30_coff_howto_table + 6;
         break;

      case BFD_RELOC_PIC30_DO_ABSOLUTE:
         return pic30_coff_howto_table + 7;
         break;

      case BFD_RELOC_PIC30_SIGNED_10_BYTE:
         return pic30_coff_howto_table + 8;
         break;

      case BFD_RELOC_PIC30_UNSIGNED_10:
         return pic30_coff_howto_table + 9;
         break;

      case BFD_RELOC_PIC30_UNSIGNED_14:
         return pic30_coff_howto_table + 10;
         break;

      case BFD_RELOC_PIC30_FRAME_SIZE:
         return pic30_coff_howto_table + 11;
         break;

      case BFD_RELOC_PIC30_WORD:
         return pic30_coff_howto_table + 12;
         break;

      case BFD_RELOC_PIC30_WORD_TBLOFFSET:
         return pic30_coff_howto_table + 13;
         break;

      case BFD_RELOC_PIC30_WORD_PSVOFFSET:
         return pic30_coff_howto_table + 14;
         break;

      case BFD_RELOC_PIC30_WORD_TBLPAGE:
         return pic30_coff_howto_table + 15;
         break;

      case BFD_RELOC_PIC30_WORD_PSVPAGE:
         return pic30_coff_howto_table + 16;
         break;

      case BFD_RELOC_PIC30_WORD_HANDLE:
         return pic30_coff_howto_table + 17;
         break;

      case BFD_RELOC_PIC30_PCREL_BRANCH:
         return pic30_coff_howto_table + 18;
         break;

      case BFD_RELOC_PIC30_BRANCH_ABSOLUTE:
         return pic30_coff_howto_table + 19;
         break;

      case BFD_RELOC_PIC30_PGM_ADDR_LSB:
         return pic30_coff_howto_table + 20;
         break;

      case BFD_RELOC_PIC30_PGM_ADDR_MSB:
         return pic30_coff_howto_table + 21;
         break;

      case BFD_RELOC_PIC30_TBLOFFSET:
         return pic30_coff_howto_table + 22;
         break;

      case BFD_RELOC_PIC30_PSVOFFSET:
         return pic30_coff_howto_table + 23;
         break;

      case BFD_RELOC_PIC30_TBLPAGE:
         return pic30_coff_howto_table + 24;
         break;

      case BFD_RELOC_PIC30_PSVPAGE:
         return pic30_coff_howto_table + 25;
         break;

      case BFD_RELOC_PIC30_HANDLE:
         return pic30_coff_howto_table + 26;
         break;

      case BFD_RELOC_PIC30_PADDR:
         return pic30_coff_howto_table + 27;
         break;

      case BFD_RELOC_PIC30_P_TBLOFFSET:
         return pic30_coff_howto_table + 28;
         break;

      case BFD_RELOC_PIC30_P_PSVOFFSET:
         return pic30_coff_howto_table + 29;
         break;

      case BFD_RELOC_PIC30_P_TBLPAGE:
         return pic30_coff_howto_table + 30;
         break;

      case BFD_RELOC_PIC30_P_PSVPAGE:
         return pic30_coff_howto_table + 31;
         break;

      case BFD_RELOC_PIC30_P_HANDLE:
         return pic30_coff_howto_table + 32;
         break;

      case BFD_RELOC_PIC30_P_PADDR:
         return pic30_coff_howto_table + 33;
         break;

      case BFD_RELOC_PIC30_BIT_SELECT_3:
         return pic30_coff_howto_table + 34;
         break;

      case BFD_RELOC_PIC30_BIT_SELECT_4_BYTE:
         return pic30_coff_howto_table + 35;
         break;

      case BFD_RELOC_PIC30_BIT_SELECT_4:
         return pic30_coff_howto_table + 36;
         break;

      case BFD_RELOC_PIC30_DSP_6:
         return pic30_coff_howto_table + 37;
         break;

      case BFD_RELOC_PIC30_DSP_PRESHIFT:
         return pic30_coff_howto_table + 38;
         break;

      case BFD_RELOC_PIC30_PWRSAV_MODE:
         return pic30_coff_howto_table + 39;
         break;

      case BFD_RELOC_PIC30_PBYTE:
         return pic30_coff_howto_table + 40;
         break;

      case BFD_RELOC_PIC30_PWORD:
         return pic30_coff_howto_table + 41;
         break;

      case BFD_RELOC_8:
         return pic30_coff_howto_table + 42;
         break;

      case BFD_RELOC_16:
         return pic30_coff_howto_table + 43;
         break;

      case BFD_RELOC_32:
         return pic30_coff_howto_table + 44;
         break;

#ifdef BFD64
      case BFD_RELOC_64:
         return pic30_coff_howto_table + 45;
         break;
#endif

      default:
         BFD_FAIL();

         return NULL;
         break;
   } /* switch (code) */
} /* static reloc_howto_type * pic30_bfd_reloc_type_lookup(...) */
#define coff_bfd_reloc_type_lookup pic30_bfd_reloc_type_lookup

#define BADMAG(x) PIC30BADMAG(x)

/* Compute the addend of a reloc.  If the reloc is to a common symbol,
   the object file contains the value of the common symbol.  By the
   time this is called, the linker may be using a different symbol
   from a different object file with a different value.  Therefore, we
   hack wildly to locate the original symbol from this file so that we
   can make the correct adjustment.  This macro sets coffsym to the
   symbol from the original file, and uses it to set the addend value
   correctly.  If this is not a common symbol, the usual addend
   calculation is done, except that an additional tweak is needed for
   PC relative relocs.
   FIXME: This macro refers to symbols and asect; these are from the
   calling function, not the macro arguments.  */
#define CALC_ADDEND(abfd, ptr, reloc, cache_ptr)                        \
  {                                                                     \
    coff_symbol_type * coffsym = (coff_symbol_type *) NULL;             \
                                                                        \
    if (ptr && bfd_asymbol_bfd (ptr) != abfd)                           \
      coffsym = (obj_symbols (abfd)                                     \
                 + (cache_ptr->sym_ptr_ptr - symbols));                 \
    else if (ptr)                                                       \
      coffsym = coff_symbol_from (abfd, ptr);                           \
                                                                        \
    if (coffsym != (coff_symbol_type *) NULL                            \
        && coffsym->native->u.syment.n_scnum == 0)                      \
      cache_ptr->addend = - coffsym->native->u.syment.n_value;          \
    else if (ptr && bfd_asymbol_bfd (ptr) == abfd                       \
             && (ptr->section != (asection *) NULL))                    \
      cache_ptr->addend = - (ptr->section->vma + ptr->value);           \
    else                                                                \
      cache_ptr->addend = 0;                                 \
                                                                        \
    if ((ptr) && (!pic30_adjustable_against_section(reloc.r_type)))     \
      cache_ptr->addend = 0;                                 \
  }


/*
 * Over-ride some of the generic coff entry points
 *
 */
#define coff_bfd_final_link pic30_final_link
#define coff_set_section_contents pic30_set_section_contents
#include "coffcode.h"


static void
bfd_pic30_print_section_header()
{
  printf("  %8s %5s %6s %6s %6s\n",
         " section", "flags", "   VMA", "   LMA", "length");
  printf("  %8s %4s %6s %6s %6s\n",
         " -------", "-----", "------", "------", "------");
}

/*
** This routine is called via bfd_map_over_sections().
*/
static void
bfd_pic30_report_sections(abfd, sect, fp) 
     bfd *abfd ATTRIBUTE_UNUSED ;
     asection *sect;
     PTR fp ATTRIBUTE_UNUSED ;
{
  printf("  %8s %5x %6x %6x %6x\n", sect->name, sect->flags,
         (unsigned int) sect->vma,
         (unsigned int) sect->lma,
         (unsigned int) sect->_raw_size);
  return;
}

/*
** Write a data record header
*/
static void
bfd_pic30_write_data_header(d, addr, len, format)
     char **d;
     bfd_vma addr;
     bfd_vma len;
     int format;
{
  /* write destination address */
  *(*d)++ = (unsigned char) (addr & 0xFF);
  *(*d)++ = (unsigned char) (addr >> 8);
  *(*d)++ = 0; *(*d)++ = 0; /* skip upper and phantom bytes */

  /* write destination length */
  *(*d)++ = (unsigned char) (len & 0xFF);
  *(*d)++ = (unsigned char) (len >> 8);
  *(*d)++ = 0; *(*d)++ = 0; /* skip upper and phantom bytes */

  /* write format code */
  *(*d)++ = (unsigned char) (format & 0xFF);
  *(*d)++ = (unsigned char) (format >> 8);
  *(*d)++ = 0; *(*d)++ = 0; /* skip upper and phantom bytes */
}


/*
** This routine is called via bfd_map_over_sections().
**
** If the section is BSS or DATA, write the appropriate
** record into the data init template.
*/
static void
bfd_pic30_process_data_sections(abfd, sect, fp) 
     bfd *abfd ATTRIBUTE_UNUSED ;
     asection *sect;
     PTR fp;
{
  unsigned char *buf,*p;
  unsigned char **d = (unsigned char **) fp;
  bfd_vma runtime_size = sect->_raw_size / 2; /* skip phantoms at runtime */

  enum { CLEAR, COPY2, COPY3 }; /* note matching definition in crt0.c */

  /* skip any persistent data section */
  if (strstr(sect->name, ".pbss") != 0)
    {
      if (pic30_debug)
        printf("  %s (skipped), size = %x\n",
               sect->name, (unsigned int) runtime_size);
      return;
    }

  /* process BSS-type sections */
  if (((sect->flags & (SEC_ALLOC | SEC_LOAD | SEC_DATA)) == SEC_ALLOC) &&
      (sect->_raw_size > 0))
    {
      if (pic30_debug)
        printf("  %s (bss), size = %x bytes, template addr = %lx\n",
               sect->name, (unsigned int) runtime_size,
               (long unsigned int) *d);

      /* write header only */
      bfd_pic30_write_data_header(d, sect->vma, runtime_size, CLEAR);
    }

  /* process DATA-type sections */
  if (((sect->flags & SEC_DATA) == SEC_DATA) && (sect->_raw_size > 0))
    {
      if (pic30_debug)
        printf("  %s (data), size = %x bytes, template addr = %lx\n",
               sect->name, (unsigned int) runtime_size,
               (long unsigned int) *d);

      /* load a copy of the section contents */
      buf = (unsigned char *) malloc (sect->_raw_size);
      if (!buf)
        {
          fprintf( stderr, "Link Error: not enough memory for section contents\n");
          abort();
        }
      if (!bfd_get_section_contents(abfd, sect, buf, 0, sect->_raw_size))
        {
          fprintf( stderr, "Link Error: can't load section %s contents\n",
                   sect->name);
          abort();
        }

      if (pic30_pack_data)
        { int count = 0;
          /* write header */
          bfd_pic30_write_data_header(d, sect->vma, runtime_size, COPY3);
          /* write packed data */
          /*   _raw_size is a multiple of 4, yet this loop wants to read
               six bytes (3 data bytes, and 3 phantom bytes) at a time.
               We may read 2 bytes (1 data and 1 phantom) to fill up the
               output record piece-meal. */
          for (p = buf; p < (buf + sect->_raw_size); )
            {
              *(*d)++ = *p++; p++; /* copy data bytes, avoid phantom */
	      count++;
              if (count == 3) {
                 *(*d)++ = 0;         /* skip phantom byte in template  */
                 count=0;
              }
            }
          /* fill template with zeroes */
          if (count) while (count < 4) { *(*d)++ = 0; count++; }
        }
      else /* (!pic30_pack_data) */
        {
          /* write header */
          bfd_pic30_write_data_header(d, sect->vma, runtime_size, COPY2);
          /* write non-packed data */
          for (p = buf; p < (buf + sect->_raw_size); )
            {
              *(*d)++ = *p++; p++; /* copy data bytes, avoid phantom */
              *(*d)++ = *p++; p++;
              *(*d)++ = DATA_UPPER_FILL_BYTE;
              *(*d)++ = 0;         /* skip upper, phantom bytes in template*/ 
            }
        }

      free(buf);

    } /* process DATA-type sections */

  return;
} /* static void bfd_pic30_process_data_sections (...)*/


/*****************************************************************************/

/*
** Call the generic final link routine, then scan
** all of the initialized data sections and write
** data init records into section .dinit.
*/
static boolean
pic30_final_link (abfd, info)
     bfd *abfd;
     struct bfd_link_info *info;
{
  boolean result;
  asection *sec;
  unsigned char *dat;

  /* call the generic final link routine */
  if (pic30_debug)
    {
      printf("\nBefore generic final link:\n");
      bfd_pic30_print_section_header();
      bfd_map_over_sections(abfd, &bfd_pic30_report_sections, 0);
    }

  result = _bfd_generic_final_link (abfd, info);

  if (pic30_debug)
    {
      printf("\nAfter generic final link:\n");
      bfd_pic30_print_section_header();
      bfd_map_over_sections(abfd, &bfd_pic30_report_sections, 0);
    }

  if (pic30_data_init)
    {
      /* get a pointer to the data template output section */
      sec = bfd_get_section_by_name(abfd, ".dinit");
      if (!sec)
        {
          fprintf( stderr, "Link Error: could not access data template\n");
          abort();
        }

      /* fill the template with a default value */
      init_data = memset( init_data, 0x65, sec->_raw_size);

      /* scan sections and write data records */
      if (pic30_debug)
          printf("\nProcessing data sections:\n");
      dat = init_data;
      bfd_map_over_sections(abfd, &bfd_pic30_process_data_sections, &dat);

      /* write zero terminator */
      *dat++ = 0; *dat++ = 0;
      *dat++ = 0; *dat++ = 0;      

      if (pic30_debug)
          printf("  last template addr written = %lx\n",
                 (long unsigned int) dat - 1);

      /* insert buffer into the data template section */
      bfd_set_section_contents (abfd, sec, init_data, 0, sec->_raw_size);
    }

#if 0
  if (pic30_debug)
    {
      printf("\nAfter pic30 final link:\n");
      bfd_pic30_print_section_header();
      bfd_map_over_sections(abfd, &bfd_pic30_report_sections, 0);
    }
#endif

  return result;
} /* static boolean pic30_final_link (...)*/


/*****************************************************************************/

/*
 *  This routine is cloned from coff_set_section_contents() in coffcode.h
 *  and modifed to support the following pic30-specific need:
 *
 *     READONLY data sections are expanded as they are written
 *     so that the program memory image includes space for
 *     the upper byte and phantom byte. Note that the output
 *     section size is already correct -- phantom bytes in data
 *     memory have reserved space for the upper and phantom
 *     byte in program memory.
 *
 */
static boolean
pic30_set_section_contents (abfd, section, location, offset, count)
     bfd *abfd;
     sec_ptr section;
     PTR location;
     file_ptr offset;
     bfd_size_type count;
{
  PTR pic30_location = location;
  file_ptr pic30_offset = offset;
  bfd_size_type pic30_count = count;
  flagword section_flags = bfd_get_section_flags (abfd, section);

#if 0
  if (pic30_debug)
    printf("set_section_contents: %s %x %x %x\n", section->name, section_flags,
           (unsigned int) section->vma, (unsigned int) section->_raw_size);
#endif

  /* If this is a linker pass (lma > 0) on
     a READONLY data section (SEC_READONLY)
     then expand the section to include padding for
     the extra width of program memory.

     Since data sections are always aligned we know
     the length is even.

     Because of PC addressing, the size of the output
     section is already correct. Basically we are
     swapping 2 phantom bytes in data memory for 2
     padding bytes in program memory. */
  if ((section->lma > 0) && (section_flags & SEC_READONLY))
    {
      /* expand data image for program memory storage */
      unsigned int j = 0;
      unsigned char *p1 = location;
      unsigned char *p2 = bfd_malloc(count);
      if (!p2) return false;
      while (j < count)
        {
          /* copy two bytes of data, then two byte of padding */
          /* (convert data phantoms into program mem padding) */
          p2[j++] = *p1++; p1++;
          p2[j++] = *p1++; p1++;
          p2[j++] = 0;
          p2[j++] = 0;
#if 0
          printf(" %lx", *((unsigned long *)(p2+j-4)));
#endif
        }
      pic30_location = p2;
      pic30_offset = offset;
      pic30_count  = count;
      if (pic30_debug)
        printf("expanding %#x bytes into %s at %#lx\n",
               (unsigned int) pic30_count, section->name, pic30_offset);
    }

  if (abfd->output_has_begun == false)	/* set by bfd.c handler */
    {
      if (! coff_compute_section_file_positions (abfd))
        {
          if (pic30_debug)
            printf("coff_compute_section_file_positions() failed!\n");
          return false;
        }
    }

  /* Don't write out bss sections - one way to do this is to
       see if the filepos has not been set.  */
  if (section->filepos == 0)
    return true;

  if (bfd_seek (abfd, (file_ptr) (section->filepos + pic30_offset), SEEK_SET) != 0)
    {
      if (pic30_debug)
        printf("bfd_seek() failed!\n");
      return false;
    }

  if (count != 0)
    {
      return (bfd_write (pic30_location, 1, pic30_count, abfd) == pic30_count) ?
        true : false;
    }
  return true;
}


/*****************************************************************************/

/*
 * Defines a little-endian bfd_target structure for coff-pic30 named
 * pic30_coff_vec.
 * There are no EXTRA_O_FLAGS (object flags) or EXTRA_S_FLAGS (section flags).
 * The character normally found at the front of a symbol is '_'.  There is no
 * data needed by back-end routines.
 */
CREATE_LITTLE_COFF_TARGET_VEC (pic30_coff_vec, "coff-pic30", 0, 0, '_', NULL)

/*****************************************************************************/

extern bfd_reloc_status_type
pic30_bfd_reloc_range_check (howto, relocation, error_msg)
  reloc_howto_type *howto;
  bfd_vma relocation;
  char ** error_msg;

/*****************************************************************************
 *
 * Special function for range checking relocation types that BFD can't handle
 *
 * Called by bfd_perform_relocation() in bfd/reloc.c when the HOWTO specifies
 * "complain_overflow_dont".
 *
 * Note that we can't use the range check macros defined in coff-pic30.h
 * because relocation is an unsigned type.
 *
 *****************************************************************************/

{
  bfd_reloc_status_type rc =  bfd_reloc_ok;

   switch (howto->type)
   {
      case R_PIC30_PBYTE:
      case R_PIC30_SIGNED_10_BYTE:
      case R_RELBYTE:
        /* valid range is [-128..255] */
        if ((relocation > 0xFF) && ~(relocation | 0x7F))
          rc = bfd_reloc_overflow;
        break;
      case R_PIC30_WORD:
      case R_RELWORD:
        /* valid range is [-32768..65535] */
        if ((relocation > 0xFFFF) && ~(relocation | 0x7FFF))
          rc = bfd_reloc_overflow;
        break;
      case R_PIC30_PWORD:
        /* valid range is [--8388608..16777215] */
        if ((relocation > 0xFFFFFF) && ~(relocation | 0x7FFFFF))
          rc = bfd_reloc_overflow;
        break;
      case R_PIC30_PCREL_DO:
      case R_PIC30_DO_ABSOLUTE:
        /* valid range is [-32768..32767] and not [-2, -1, 0] */
        if ((relocation > 0x7FFF) && ~(relocation | 0xC0007FFF))
          rc = bfd_reloc_overflow;
        if (((relocation & 0xFFFF) == 0xFFFE) ||
            ((relocation & 0xFFFF) == 0xFFFF) ||
            (relocation == 0))
          {
            *error_msg = (char *) malloc(BUFSIZ);
            sprintf(*error_msg,
                 "Illegal value for DO instruction offset (-2, -1, or 0)");
            rc = bfd_reloc_dangerous;
          }
        break;
      case R_PIC30_DSP_6:
        /* valid range is [-16..16] */
        if ((relocation > 0x10) && ~(relocation | 0xF))
          rc = bfd_reloc_overflow;
        break;
   }

   return rc;
} /* extern bfd_reloc_status_type pic30_bfd_reloc_range_check (...)*/


/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_is_even_address (abfd, reloc_entry, symbol, data, input_section,
                            output_bfd, error_msg)
   bfd * abfd ATTRIBUTE_UNUSED;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data ATTRIBUTE_UNUSED;
   asection * input_section ATTRIBUTE_UNUSED;
   bfd * output_bfd ATTRIBUTE_UNUSED;
   char ** error_msg;

/*****************************************************************************
 *
 * Special function for ensuring that the symbol being relocated is located at
 * an even address.
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Target specific assembler functions have already checked the
 *             bits.  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  If the address is ODD, generate an error message, and return
 *             DANGEROUS; otherwise, call the generic special function to
 *             finish howto processing.  NOTE:  We do not include the
 *             information that exists in the instruction as the assembler has
 *             already verified that it is even.
 *                An even + an odd => an odd.
 *                An even + an even => an even.
 *
 *****************************************************************************/

{
   bfd_reloc_status_type rc;

   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       rc = bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         rc = bfd_reloc_undefined;
      else if (!PIC30_IS_EVEN (symbol->value + reloc_entry->addend))
      {
         *error_msg = (char *) malloc(BUFSIZ);
         sprintf(*error_msg,
                 "Cannot access symbol (%s) at an odd address.", symbol->name);
         rc = bfd_reloc_dangerous;
      } /* if (!PIC30_IS_EVEN (symbol->value + reloc_entry->addend)) */
      else
        /* finish by calling the generic special function */
        rc = pic30_coff_perform_generic(abfd, reloc_entry, symbol, data,
                                        input_section, output_bfd, error_msg);

   } /* else */

   return rc;
} /* static bfd_reloc_status_type pic30_coff_is_even_address(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_is_file_reg_address (abfd, reloc_entry, symbol, data, input_section,
                                output_bfd, error_msg)
   bfd * abfd ATTRIBUTE_UNUSED;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data ATTRIBUTE_UNUSED;
   asection * input_section ATTRIBUTE_UNUSED;
   bfd * output_bfd ATTRIBUTE_UNUSED;
   char ** error_msg;

/*****************************************************************************
 *
 * Special function for ensuring that the FILE_REG and FILE_REG_BYTE
 * relocations fall within the file register address space.
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Target specific assembler functions have already checked the
 *             bits.  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  Pre-calculate the relocation. If the result is >= FILE_REG_LIMIT,
 *             generate an error message, and return DANGEROUS.
 *             Otherwise continue the howto processing as follows:
 *
 *             if howto->type == R_PIC30_FILE_REG
 *               call pic30_coff_is_even_address()
 *             else
 *               call pic30_coff_perform_generic()
 *
 *****************************************************************************/

{
   bfd_reloc_status_type rc;

   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       rc = bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      reloc_howto_type * howto = reloc_entry->howto;
      bfd_size_type octets = reloc_entry->address;
      bfd_size_type offset;
      bfd_vma relocation;

      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
        return bfd_reloc_undefined;  /* exit if symbol is undefined */

      /* pre-calculate the relocation */
      offset = pic30_coff_extract_bytes (abfd, data, 4, octets, true, true);
      offset &= howto->src_mask;
      offset >>= howto->bitpos;

      offset += symbol->section->output_offset
        + reloc_entry->addend;

      relocation = symbol->value
        + symbol->section->output_section->vma
        + offset;

      relocation >>= (bfd_vma) howto->rightshift;

      /* range check the final value */
      if (relocation >= FILE_REG_LIMIT)
        {
          char *c1 = "Cannot access symbol ";
          char *c2 = "with file register addressing. Value must be less than 8192.";
          *error_msg = (char *) malloc(BUFSIZ);

          if (offset)
            sprintf(*error_msg, "%s(%s+%lu) %s", c1, symbol->name, offset, c2);
          else
            sprintf(*error_msg, "%s(%s) %s", c1, symbol->name, c2);
          return bfd_reloc_dangerous;  /* exit on overflow */
        }

      if (howto->type == R_PIC30_FILE_REG)
        /* finish by calling the even address special function */
        rc = pic30_coff_is_even_address(abfd, reloc_entry, symbol, data,
                                        input_section, output_bfd, error_msg);
      else
        /* finish by calling the generic special function */
        rc = pic30_coff_perform_generic(abfd, reloc_entry, symbol, data,
                                        input_section, output_bfd, error_msg);

   } /* else */

   return rc;
} /* static bfd_reloc_status_type pic30_coff_is_file_reg_address(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_program_address (abfd, reloc_entry, symbol, data,
                                    input_section, output_bfd, error_msg)
   bfd * abfd ATTRIBUTE_UNUSED;
   arelent * reloc_entry ATTRIBUTE_UNUSED;
   asymbol * symbol;
   PTR data ATTRIBUTE_UNUSED;
   asection * input_section ATTRIBUTE_UNUSED;
   bfd * output_bfd;
   char ** error_msg;

/*****************************************************************************
 *
 * Special function for ensuring that the symbol being relocated is located
 * in either a CODE or ABSOLUTE section.
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  If the address is not in either a code or the absolute section,
 *             generate an error message, and return DANGEROUS; otherwise,
 *             call generic special function to finish the howto processing.
 *             Undefined weak symbols are acceptable; the generic special
 *             function will relocate them with a value of zero.
 *
 *****************************************************************************/

{
   bfd_reloc_status_type rc;

   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       rc = bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         rc = bfd_reloc_undefined;
      else if (((symbol->section->flags & SEC_CODE) == 0) &&
               (!bfd_is_abs_section (symbol->section) &&
                !bfd_is_und_section (symbol->section)))
      {
         *error_msg = (char *) malloc(BUFSIZ);
         sprintf(*error_msg,
                 "Symbol (%s) is not located in an executable section.",
                 symbol->name);
         rc = bfd_reloc_dangerous;
      } /* if ((symbol->section->flags & SEC_CODE) == 0) &&
               (!bfd_is_abs_section (symbol->section) &&
               !bfd_is_und_section (symbol->section))) */
      else
        /* finish by calling the generic special function */
        rc = pic30_coff_perform_generic(abfd, reloc_entry, symbol, data,
                                        input_section, output_bfd, error_msg);
   } /* else */

   return rc;
} /* static bfd_reloc_status_type pic30_coff_perform_program_address(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_pc_relative (abfd, reloc_entry, symbol, data,
                                input_section, output_bfd, error_msg)
   bfd * abfd ATTRIBUTE_UNUSED;
   arelent * reloc_entry;
   asymbol * symbol ATTRIBUTE_UNUSED;
   PTR data ATTRIBUTE_UNUSED;
   asection * input_section ATTRIBUTE_UNUSED;
   bfd * output_bfd;
   char ** error_msg ATTRIBUTE_UNUSED;

/*****************************************************************************
 *
 * Special function for processing PC-relative relocations.
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  If there is an offset in the instruction, convert it into
 *             a relocation addend and clear the instruction field.
 *             Add 4 to the relocation entry address.  PC-relative offsets
 *             are calculated from the next instruction word.  Also check to
 *             ensure that the address being branched to is even.  If not,
 *             display an error message and return DANGEROUS; otherwise,
 *             call pic30_coff_perform_generic() to finish the relocation.
 *
 * Note:  Unlike other offsets and lengths, the relocation address
 *        is provided in octets, not PC units.
 * 
 *****************************************************************************/

{
   bfd_reloc_status_type rc = bfd_reloc_continue;

   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       rc = bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         rc = bfd_reloc_undefined;
      else
      {
        /* look for instruction offset */
        bfd_size_type octets = reloc_entry->address;
        short offset = pic30_coff_extract_bytes (abfd, data, 2, octets,
                                                 true, true);

        if (offset) /* store as addend, clear instruction offset */
          {
            reloc_entry->addend = offset;
            pic30_coff_insert_bytes (abfd, data, 2, octets, 0, true, true);
          }

        /* adjust reloc address by 1 pword (4 octets) to get offset right */
        reloc_entry->address += 4;

        /* finish by calling the even_address special function */
        rc = pic30_coff_is_even_address (abfd, reloc_entry, symbol, data,
                                         input_section, output_bfd, error_msg);
      } /* else */
   } /* else */

   return rc;
} /* static bfd_reloc_status_type pic30_coff_perform_pc_relative(...) */


/*****************************************************************************
 * Utility routine for data relocations.
 *****************************************************************************/

static bfd_vma
pic30_coff_extract_bytes (abfd, data, count, addr, skip_upper, skip_phantom)
   bfd * abfd ATTRIBUTE_UNUSED;
   PTR data;
   int count;
   bfd_size_type addr;
   unsigned char skip_upper;
   unsigned char skip_phantom;

/*****************************************************************************
 *
 * Extracts COUNT bytes from DATA with ability to skip upper and phantom bytes.
 *
 *****************************************************************************/

{
   unsigned char x;
   bfd_size_type octets = addr;
   bfd_vma target = 0;
   int i;

   for (i = 0; i < count; i++)
   {
      /* If address is upper byte, skip ahead if requested */
      if (skip_upper)
         if ((octets & 0x02) == 0x02)
            octets++;

      /* If address is phantom byte, skip ahead if requested */
      if (skip_phantom)
         if ((octets & 0x03) == 0x03)
            octets++;

      /* Read a byte */
      x = bfd_get_8 (abfd, (char *) data + octets);

      /* Accumulate it */
      target = (target | (bfd_vma) (x << (8 * i)));
      octets++;
   } /* for i */

   return target;
} /* static bfd_vma pic30_coff_extract_bytes(...) */

/*****************************************************************************/

static void
pic30_coff_insert_bytes (abfd, data, count, addr, value,
                        skip_upper, skip_phantom)
   bfd * abfd ATTRIBUTE_UNUSED;
   PTR data;
   int count;
   bfd_size_type addr;
   bfd_vma value;
   unsigned char skip_upper;
   unsigned char skip_phantom;

/*****************************************************************************
 *
 * Inserts COUNT bytes of VALUE into DATA with ability to skip upper and
 * phantom bytes.
 *
 *****************************************************************************/

{
   unsigned char x;
   bfd_size_type octets = addr;
   int i;

   for (i = 0; i < count; i++)
   {
      /* If address is upper byte, skip ahead if requested */
      if (skip_upper)
         if ((octets & 0x02) == 0x02)
            octets++;

      /* If address is phantom byte, skip ahead if requested */
      if (skip_phantom)
         if ((octets & 0x03) == 0x03)
             octets++;

      /* Select a byte in the value */
      x = (value >> (i * 8) & 0xFF);

      /* Insert it */
      bfd_put_8 (abfd, x, (unsigned char *) data + octets);
      octets++;
   } /* for i */
} /* static void pic30_coff_insert_bytes(...) */

/*****************************************************************************/

static bfd_vma
pic30_coff_extract_bytes_data_mem (abfd, data, count, addr)
   bfd * abfd ATTRIBUTE_UNUSED;
   PTR data;
   int count;
   bfd_size_type addr;

/*****************************************************************************
 *
 * Extract COUNT bytes from a data memory section. Skips phantom bytes.
 *
 *****************************************************************************/

{
   unsigned char x;
   bfd_size_type octets = addr;
   bfd_vma target = 0;
   int i;

   for (i = 0; i < count; i++)
   {
      /* If address is phantom byte, skip ahead */
      if ((octets & 0x01) == 0x01)
        octets++;

      /* Read a byte */
      x = bfd_get_8 (abfd, (char *) data + octets);

      /* Accumulate it */
      target = (target | (bfd_vma) (x << (8 * i)));
      octets++;
   } /* for i */

   return target;
} /* static bfd_vma pic30_coff_extract_bytes_data_mem(...) */

/*****************************************************************************/

static void
pic30_coff_insert_bytes_data_mem (abfd, data, count, addr, value)
   bfd * abfd ATTRIBUTE_UNUSED;
   PTR data;
   int count;
   bfd_size_type addr;
   bfd_vma value;

/*****************************************************************************
 *
 * Insert COUNT bytes of VALUE into a data memory section. Skips phantom bytes.
 *
 *****************************************************************************/

{
   unsigned char x;
   bfd_size_type octets = addr;
   int i;

   for (i = 0; i < count; i++)
   {
      if ((octets & 0x01) == 0x01)
          octets++;

      /* Select a byte in the value */
      x = (value >> (i * 8) & 0xFF);

      /* Insert it */
      bfd_put_8 (abfd, x, (unsigned char *) data + octets);
      octets++;
   } /* for i */
} /* static void pic30_coff_insert_bytes_data_mem(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_data_directive (abfd, reloc_entry, symbol, data,
                                   input_section, output_bfd, error_msg)
   bfd * abfd;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data;
   asection * input_section;
   bfd * output_bfd;
   char ** error_msg ATTRIBUTE_UNUSED;

/*****************************************************************************
 *
 * Special function for data directive relocations (.byte, .word, .long).
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  This special function is used in both the CODE and DATA sections
 *             for symbols located in both the CODE and DATA sections.
 *
 *             If the relocation site is in a CODE section, then the data
 *             placed there by the assembler will skip the upper and phantom
 *             bytes.  Extract the data to add to the relocation, skipping the
 *             upper and phantom bytes if necessary.  Then after checking for
 *             overflow, insert the calculated relocation back into the
 *             relocation site, skipping the upper and phantom bytes if
 *             necessary.
 *
 * Note:  Unlike other offsets and lengths, the relocation address
 *        is provided in octets, not PC units.
 *
 *****************************************************************************/

{
   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       return bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      bfd_vma relocation,offset;
      bfd_vma output_base = 0;
      bfd_size_type octets = reloc_entry->address;
      bfd_reloc_status_type flag = bfd_reloc_ok;
      reloc_howto_type * howto = reloc_entry->howto;
      asection * reloc_target_output_section;
      int count = (howto->size == 0) ? 1 : (2 * howto->size);

      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         return bfd_reloc_undefined;

      /* Get symbol value.  (Common symbols are special.)  */
      if (bfd_is_com_section (symbol->section))
         relocation = 0;
      else
         relocation = symbol->value;

      reloc_target_output_section = symbol->section->output_section;
      output_base = reloc_target_output_section->vma;

#if 0
      /* DEBUG */
      printf ("symbol->name = %s\n", symbol->name);
      printf ("symbol->value = 0x%x\n", symbol->value);
      printf ("output_base = 0x%x\n", output_base);
      printf ("reloc_target_output_section->lma = 0x%x\n",
              reloc_target_output_section->lma);
      printf ("symbol->section->output_offset = 0x%x\n",
              symbol->section->output_offset);
#endif

      relocation += symbol->section->output_offset;

      /* Add in supplied addend. */
      relocation += reloc_entry->addend;

      /* Add in the base of the mapped section. */
      relocation += output_base; 

      relocation >>= (bfd_vma) howto->rightshift;

#if 0
      printf("relocation = 0x%x\n", (long) relocation);
#endif

      /* Extract offset data from the encoded instruction. */
      if (PIC30_SECTION_IN_PROGRAM_MEMORY (input_section->output_section))
         /* Relocation site is program memory, skip upper & phantom bytes */
         offset = pic30_coff_extract_bytes (abfd, data, count, octets,
                                           true, true);
      else
         /* Relocation site is data memory */
         offset = pic30_coff_extract_bytes_data_mem (abfd, data, count, octets);

#if 0
      printf ("offset = %x\n", offset);
#endif

      /* Add in the offset */
      relocation += offset;

      /* Check final relocation for overflow */
      if (howto->complain_on_overflow != complain_overflow_dont)
         flag = bfd_check_overflow (howto->complain_on_overflow,
			            howto->bitsize,
			            0, /* rightshift already applied */
			            bfd_arch_bits_per_address (abfd),
			            relocation);
      else
        flag = pic30_bfd_reloc_range_check (howto, relocation, error_msg);

      /* Insert the final relocation into the BFD */
      if (PIC30_SECTION_IN_PROGRAM_MEMORY (input_section->output_section))
         /* Relocation site is program memory, skip upper & phantom bytes */
         pic30_coff_insert_bytes (abfd, data, count, octets, relocation,
                                  true, true);
      else
         /* Relocation site is data memory */
         pic30_coff_insert_bytes_data_mem (abfd, data, count, octets, relocation);

      /*
       * Relocation is complete.  Return the overflow flag to
       * bfd_perform_relocation
       */
      return flag;
   } /* else */
} /* static bfd_reloc_status_type pic30_coff_perform_data_directive(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_p_data_directive (abfd, reloc_entry, symbol, data,
                                     input_section, output_bfd, error_msg)
   bfd * abfd;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data;
   asection * input_section;
   bfd * output_bfd;
   char ** error_msg ATTRIBUTE_UNUSED;

/*****************************************************************************
 *
 * Special function for p-variant data directive relocations (.pbyte, .pword).
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  This special function is used in the CODE section
 *             for symbols located in both the CODE and DATA sections.  If
 *             this special function is called in a DATA section, then an
 *             error occurred in the assembler, and return NOTSUPPORTED.
 *             If the relocation site is in the CODE section, then the data
 *             placed there by the assembler will skip the phantom byte.
 *             Extract the data to add to the relocation skipping the
 *             phantom byte if necessary.  Then after checking for
 *             overflow, insert the calculated relocation back into the
 *             relocation site skipping the phantom byte if necessary.
 *
 * Note:  Unlike other offsets and lengths, the relocation address
 *        is provided in octets, not PC units.
 *
 *****************************************************************************/

{
   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       return bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      bfd_vma relocation, offset;
      bfd_vma output_base = 0;
      bfd_size_type octets = reloc_entry->address ;
      bfd_reloc_status_type flag = bfd_reloc_ok;
      reloc_howto_type * howto = reloc_entry->howto;
      asection * reloc_target_output_section;
      int count = howto->bitsize / 8;

      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         return bfd_reloc_undefined;

      /* If the relocation site is not in a CODE section, then this relocation
         is not supported.  Linker will abort if this ever occurs.  This means
         that there is a problem in the assembler! */
      if ((input_section->output_section->flags & SEC_CODE) == 0)
         return bfd_reloc_notsupported;   /* exit if not a CODE section */

      /* Get symbol value.  (Common symbols are special.)  */
      if (bfd_is_com_section (symbol->section))
         relocation = 0;
      else
         relocation = symbol->value;

      reloc_target_output_section = symbol->section->output_section;
      output_base = reloc_target_output_section->vma;

#if 0
      /* DEBUG */
      printf ("symbol->name = %s\n", symbol->name);
      printf ("symbol->value = 0x%x\n", symbol->value);
      printf ("output_base = 0x%x\n", output_base);
      printf ("reloc_target_output_section->lma = 0x%x\n",
              reloc_target_output_section->lma);
      printf ("symbol->section->output_offset = 0x%x\n",
              symbol->section->output_offset);
#endif

      relocation += symbol->section->output_offset;

      /* Add in supplied addend.  */
      relocation += reloc_entry->addend;

      /* Add in the base of the mapped section. */
      relocation += output_base; 

      relocation >>= (bfd_vma) howto->rightshift;

      /* Extract offset data from the encoded instruction, skipping phantom
         byte */
      offset = pic30_coff_extract_bytes (abfd, data, count, octets,
                                         false, true);

#if 0
      printf("offset = %x\n",offset);
      printf("output_base = %lx\n",output_base);
#endif

      /* Add in the offset */
      relocation += offset;

#if 0
      printf("relocation = %x\n",relocation);
#endif

      /* check final relocation for overflow */
      if (howto->complain_on_overflow != complain_overflow_dont)
         flag = bfd_check_overflow (howto->complain_on_overflow,
			            howto->bitsize,
			            0, /* rightshift already applied */
			            bfd_arch_bits_per_address (abfd),
			            relocation);
      else
        flag = pic30_bfd_reloc_range_check (howto, relocation, error_msg);
             

      /* Insert the final relocation into the instruction, skipping phantom
         byte */
#if 0
      printf("octets = %lx\n", octets);
      printf("count = %u\n", count);
#endif
      pic30_coff_insert_bytes (abfd, data, count, octets, relocation,
                               false, true);

      /*
       * Relocation is complete.  Return the overflow flag to
       * bfd_perform_relocation
       */
      return flag;
   } /* else */
} /*  static bfd_reloc_status_type pic30_coff_perform_p_data_directive(...) */

/*****************************************************************************/

static bfd_vma
bfd_pic30_handle (abfd, addr, symbol)
     bfd *abfd;
     bfd_vma addr;
     asymbol *symbol;

/*****************************************************************************
 *
 * Resolve handle relocations by looking up the symbol in the handle hash
 * table and returning the (16 bit) jump table entry.
 *
 * Note: if address fits in 16 bits, just return it.
 *
 *****************************************************************************/

{
  bfd_vma target;

#if 0
  printf("resolving handle to %s\t%x\n", symbol->name, addr);
#endif

  if ((addr == (addr & 0xFFFF)) ||  /* if address fits in 16 bits      */
       (!handles))                  /*  or no handle hash table exists */
    return addr;
  else
    {
      struct handle_hash_entry *handle;
      asection *sec;
      char *name;
      bfd_size_type offset;
      bfd_vma table;

      /* if symbol is local, construct a global name */
      if ((symbol->flags & BSF_LOCAL) == BSF_LOCAL)
        {
          name = (char *) bfd_malloc ( strlen(abfd->filename) +
                                       strlen(symbol->name) + 2);
          if (!name)
            {
              fprintf( stderr, 
                       "Link Error: not enough memory for handle symbol\n");
              abort();
            }
          strcpy (name, abfd->filename);
          strcat (name, ":");
          strcat (name, symbol->name);
        }
      else
        name = (char *) symbol->name;

      /* lookup the handle hash entry for this symbol */
#if 0
      printf("looking up hash for %s\n", name);
#endif
      handle = handle_hash_lookup (handles, name, 0, 0);
      if (!handle)
        {
          fprintf( stderr, "Link Error: can't access handle %s\n", name);
          abort();
        }

      /* get a fresh copy of the handle section */
      sec = bfd_get_section_by_name(handle_bfd, ".handle");
#if 0
      printf("%s\t%lx\t%x\t%s\t%lx\n", sec->name,
             (unsigned long) sec,
             (unsigned int) sec->output_offset,
             sec->output_section->name,
             sec->output_section->vma);
#endif

      /* get the base address of the jump table */
      table = sec->output_section->vma + sec->output_offset;

      /* calculate the offset of this entry     */
      offset = handle->index * 4;

      /* sanity check the result */
      target = table + offset;
      if (target != (target & 0xFFFF))
          fprintf( stderr, "Linker Script Error: section .handle ",
                   "must be allocated low in program memory.\n");
      return (target);
        
    }
}

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_operators (abfd, reloc_entry, symbol, data,
                              input_section, output_bfd, error_msg)
   bfd * abfd;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data;
   asection * input_section;
   bfd * output_bfd;
   char ** error_msg;

/*****************************************************************************
 *
 * Special function for tbloffset(), psvoffset(), tblpage(), psvpage(),
 * handle(), paddr() operators.
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  This special function is used in both the CODE and DATA sections
 *             for symbols located in a CODE or READONLY section. If this
 *             special function is called to relocate a symbol in a DATA
 *             section, then an error message is generated and DANGEROUS
 *             is returned. Undefined weak symbols are acceptable; they
 *             resolve to a value of zero.
 *
 *             If the relocation site is in a CODE section and a data
 *             allocation directive was used, then the data placed there by
 *             the assembler will skip the upper and phantom bytes.  Extract
 *             the data to add to the relocation skipping the upper and
 *             phantom bytes if necessary.  Add this value to the relocation
 *             and then apply the operator.  Then after checking for overflow,
 *             insert the calculated relocation back into the relocation site
 *             skipping the upper and phantom bytes if necessary.
 * 
 * Note1:  This is a 32-bit operation because the reloc may not be
 *         right-justified in the instruction (howto->bitpos > 0).
 *
 * Note2:  Unlike other offsets and lengths, the relocation address
 *         is provided in octets, not PC units.
 *
 *****************************************************************************/

{
   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       return bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      bfd_vma relocation, target;
      bfd_vma output_base = 0;
      bfd_size_type octets = reloc_entry->address;
      bfd_reloc_status_type flag = bfd_reloc_ok;
      reloc_howto_type * howto = reloc_entry->howto;
      asection * reloc_target_output_section;
      unsigned long x;

      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         return bfd_reloc_undefined;

      if (((symbol->section->flags & (SEC_CODE | SEC_READONLY)) == 0) &&
          !bfd_is_und_section (symbol->section))
      {
         *error_msg = (char *) malloc(BUFSIZ);
         sprintf(*error_msg, "Cannot use %s on a symbol (%s) that is not "
                             "located in an executable or readonly section.",
                 howto->name, symbol->name);
         return bfd_reloc_dangerous;
      } /* if ((symbol->section->flags & (SEC_CODE | SEC_READONLY)) == 0) &&
               !bfd_is_und_section (symbol->section))*/


      /* Get symbol value.
         Common and undefined weak symbols resolve to zero. */
      if (bfd_is_com_section (symbol->section) ||
          (bfd_is_und_section (symbol->section) && (symbol->flags & BSF_WEAK)))
         relocation = 0;
      else
         relocation = symbol->value;

      reloc_target_output_section = symbol->section->output_section;

      /* Convert input-section-relative symbol value to absolute.  */
      output_base = reloc_target_output_section->lma;

#if 0
      /* DEBUG */
      printf ("symbol->name = %s\n", symbol->name);
      printf ("symbol->value = 0x%x\n", symbol->value);
      printf ("output_base = 0x%x\n", output_base);
      printf ("symbol->section->output_offset = 0x%x\n",
              symbol->section->output_offset);
      printf ("reloc_entry->addend = %#x\n", reloc_entry->addend);
#endif

      relocation += output_base + symbol->section->output_offset;

      /* Add in supplied addend.  */
      relocation += reloc_entry->addend;

      /* resolve handle */
      if ((howto->type == R_PIC30_WORD_HANDLE) ||
          (howto->type == R_PIC30_HANDLE))
        relocation = bfd_pic30_handle (abfd, relocation, symbol);
      
      relocation >>= (bfd_vma) howto->rightshift;

      /* Add in data from the current program word */
      if ((howto->type == R_PIC30_WORD_TBLOFFSET) ||
          (howto->type == R_PIC30_WORD_PSVOFFSET) ||
          (howto->type == R_PIC30_WORD_TBLPAGE) ||
          (howto->type == R_PIC30_WORD_PSVPAGE) ||
          (howto->type == R_PIC30_WORD_HANDLE))
      { 
         /* If instruction, do it the easy way */
         x = bfd_get_32 (abfd, (bfd_byte *) data + octets);
         target = ((x & howto->src_mask) >> howto->bitpos);
      } /* if */
      else
      {
         int count = (howto->size == 0) ? 1 : (2 * howto->size);

         if (PIC30_SECTION_IN_PROGRAM_MEMORY (input_section->output_section))
           /* Relocation site is program memory, skip upper & phantom bytes */
            target = pic30_coff_extract_bytes (abfd, data, count, octets,
                                               true, true);
         else
           /* Relocation site is data memory */
            target = pic30_coff_extract_bytes_data_mem (abfd, data, count, octets);
      } /* else */

      relocation += target;

      if ((howto->type == R_PIC30_TBLOFFSET) ||
          (howto->type == R_PIC30_WORD_TBLOFFSET))
         relocation = PIC30_TBLOFFSET (relocation);
      else if ((howto->type == R_PIC30_PSVOFFSET) ||
               (howto->type == R_PIC30_WORD_PSVOFFSET))
         relocation = PIC30_PSVOFFSET (relocation);

      if (howto->complain_on_overflow != complain_overflow_dont)
         flag = bfd_check_overflow (howto->complain_on_overflow,
                                    howto->bitsize,
                                    0, /* rightshift already applied */
                                    /* bfd_arch_bits_per_address (abfd), */
                                    PIC30_BITS_PER_PROGRAM_ADDRESS,
                                    relocation);

#if 0
      printf("relocation = 0x%x\n", (long) relocation);
#endif

      /* Insert the final relocation */
      if ((howto->type == R_PIC30_WORD_TBLOFFSET) ||
          (howto->type == R_PIC30_WORD_PSVOFFSET) ||
          (howto->type == R_PIC30_WORD_TBLPAGE) ||
          (howto->type == R_PIC30_WORD_PSVPAGE) ||
          (howto->type == R_PIC30_WORD_HANDLE))
      {
        /* If reloc site is an instruction, do it the easy way */
        x = bfd_get_32 (abfd, (bfd_byte *) data + octets);
        x = ((x & ~howto->dst_mask) |
             ((relocation << howto->bitpos) & howto->dst_mask));
        bfd_put_32 (abfd, x, (unsigned char *) data + octets);
      } /* if */
      else
      {
         int count = (howto->size == 0) ? 1 : (2 * howto->size);

         if (PIC30_SECTION_IN_PROGRAM_MEMORY (input_section->output_section))
           /* Relocation site is program memory, skip upper & phantom bytes */
            pic30_coff_insert_bytes (abfd, data, count, octets, relocation,
                                     true, true);
         else
           /* Relocation site is program memory, skip upper & phantom bytes */
            pic30_coff_insert_bytes_data_mem (abfd, data,
                                              count, octets, relocation);
      } /* else */

      /*
       * Relocation is complete.  Return the overflow flag to
       * bfd_perform_relocation
       */
      return flag;
   } /* else */
} /* static bfd_reloc_status_type pic30_coff_perform_operators(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_p_operators (abfd, reloc_entry, symbol, data,
                                input_section, output_bfd, error_msg)
   bfd * abfd;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data;
   asection * input_section;
   bfd * output_bfd;
   char ** error_msg;

/*****************************************************************************
 *
 * Special function for tbloffset(), psvoffset(), tblpage(), psvpage(),
 * handle(), paddr() operators that appear in the p-variant data directives.
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  This special function is used in CODE sections for symbols
 *             located in a CODE or READONLY section. If this special
 *             function is called to relocate a symbol in a DATA section, then
 *             an error message is generated and DANGEROUS is returned.
 *             Undefined weak symbols are acceptable; they resolve to
 *             a value of zero.
 *
 *             If the relocation site is in a CODE section and a data
 *             allocation directive was used, then the data placed there by
 *             the assembler will skip the phantom byte. Extract the data to
 *             add to the relocation, skipping the phantom byte if necessary.
 *             Then after checking for overflow, insert the calculated
 *             relocation back into the relocation site skipping the
 *             phantom byte if necessary.
 *
 * Note:  Unlike other offsets and lengths, the relocation address
 *        is provided in octets, not PC units.
 *
 *****************************************************************************/

{
   if (output_bfd != (bfd *) NULL)
   {
       /* ASSEMBLER */
       return bfd_reloc_ok;
   } /* if (output_bfd != (bfd *) NULL) */
   else
   {
      /* LINKER */
      bfd_vma relocation, target;
      bfd_vma output_base = 0;
      bfd_size_type octets = reloc_entry->address;
      bfd_reloc_status_type flag = bfd_reloc_ok;
      reloc_howto_type * howto = reloc_entry->howto;
      asection * reloc_target_output_section;
      int count = howto->bitsize / 8;

      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
         return bfd_reloc_undefined;

      if (((symbol->section->flags & (SEC_CODE | SEC_READONLY)) == 0) &&
          !bfd_is_und_section (symbol->section))
        {
         *error_msg = (char *) malloc(BUFSIZ);
         sprintf(*error_msg, "Cannot use %s on a symbol (%s) that is not "
                             "located in an executable or readonly section.",
                 howto->name, symbol->name);
         return bfd_reloc_dangerous;
        } /* if ((symbol->section->flags & (SEC_CODE | SEC_READONLY)) == 0) &&
                 !bfd_is_und_section (symbol->section))*/

      /* If the relocation site is not in a CODE section, then this relocation
         is not supported.  Linker will abort if this ever occurs.  This means
         that there is a problem in the assembler! */
      if ((input_section->output_section->flags & SEC_CODE) == 0)
         return bfd_reloc_notsupported;   /* exit if not a CODE section */

      /* Get symbol value.
         Common and undefined weak symbols resolve to zero. */
      if (bfd_is_com_section (symbol->section) ||
          (bfd_is_und_section (symbol->section) && (symbol->flags & BSF_WEAK)))
         relocation = 0;
      else
         relocation = symbol->value;

      reloc_target_output_section = symbol->section->output_section;

      /* Convert input-section-relative symbol value to absolute.  */
      output_base = reloc_target_output_section->lma;

#if 0
      /* DEBUG */
      printf ("symbol->name = %s\n", symbol->name);
      printf ("symbol->value = 0x%x\n", symbol->value);
      printf ("output_base = 0x%x\n", output_base);
      printf ("symbol->section->output_offset = 0x%x\n",
              symbol->section->output_offset);
      printf("reloc_entry->addend = %#x\n", reloc_entry->addend);
#endif

      relocation += output_base + symbol->section->output_offset;

      /* Add in supplied addend.  */
      relocation += reloc_entry->addend;

      /* resolve handle */
      if ((howto->type == R_PIC30_P_HANDLE))
        relocation = bfd_pic30_handle (abfd, relocation, symbol);
      
      relocation >>= (bfd_vma) howto->rightshift;

      /* Extract data byte-wise, avoiding phantom bytes */
      target = pic30_coff_extract_bytes (abfd, data, count, octets,
                                         false, true);
      relocation += target;

      if (howto->type == R_PIC30_P_TBLOFFSET)
         relocation = PIC30_TBLOFFSET (relocation);
      else if (howto->type == R_PIC30_P_PSVOFFSET)
         relocation = PIC30_PSVOFFSET (relocation);

      if (howto->complain_on_overflow != complain_overflow_dont)
         flag = bfd_check_overflow (howto->complain_on_overflow,
                                    howto->bitsize,
                                    0, /* rightshift already applied */
                                    /* bfd_arch_bits_per_address (abfd), */
                                    PIC30_BITS_PER_PROGRAM_ADDRESS,
                                    relocation);

#if 0
      printf("relocation = 0x%x\n", (long) relocation);
#endif

      /* Insert data byte-wise, avoiding phantom bytes */
      pic30_coff_insert_bytes (abfd, data, count, octets, relocation,
                               false, true);

      /*
       * Relocation is complete.  Return the overflow flag to
       * bfd_perform_relocation
       */
      return flag;
   } /* else */
} /* static bfd_reloc_status_type pic30_coff_perform_p_operators(...) */

/*****************************************************************************/

static bfd_reloc_status_type
pic30_coff_perform_generic (abfd, reloc_entry, symbol, data,
                            input_section, output_bfd, error_msg)
   bfd * abfd;
   arelent * reloc_entry;
   asymbol * symbol;
   PTR data;
   asection * input_section;
   bfd * output_bfd;
   char ** error_msg;

/*****************************************************************************
 *
 * Special function for a generic relocation.
 * 
 * This code is similar to the generic code in bfd_perform_relocation()
 * with the following exceptions:
 *
 * 1. partial relocations are not supported
 * 2. reloc_entry->address is assumed to be in "octets" units
 * 3. negative values for howto->size are ignored
 *
 * Only want to perform the special function if in the linker.
 *
 * ASSEMBLER:  Return OK as the target specific assembler functions have
 *             already filled the bits with the required information.
 *
 *    LINKER:  Perform the generic relocation operation.
 *
 *****************************************************************************/

{
  if (output_bfd != (bfd *) NULL)
    {
      /* ASSEMBLER */
      return bfd_reloc_ok;
    } /* if (output_bfd != (bfd *) NULL) */
  else
    {
      /* LINKER */
      bfd_vma relocation;
      bfd_vma output_base = 0;
      bfd_size_type octets = reloc_entry->address;
      bfd_reloc_status_type flag = bfd_reloc_ok;
      reloc_howto_type * howto = reloc_entry->howto;
      asection * reloc_target_output_section;

      if ((bfd_is_und_section (symbol->section)) &&
          ((symbol->flags & BSF_WEAK) == 0))
        return bfd_reloc_undefined;

      /* only certain relocation types support CODE targets */
      if (symbol->section->flags & SEC_CODE)
        switch (howto->type)
          {
          case R_PIC30_PCREL_DO:
          case R_PIC30_DO_ABSOLUTE:
          case R_PIC30_PCREL_BRANCH:
          case R_PIC30_BRANCH_ABSOLUTE:
          case R_PIC30_PGM_ADDR_LSB:
          case R_PIC30_PGM_ADDR_MSB:
            /* these types don't use the generic special function */
            /* case R_PIC30_WORD_TBLOFFSET: */
            /* case R_PIC30_WORD_PSVOFFSET: */
            /* case R_PIC30_WORD_TBLPAGE: */
            /* case R_PIC30_WORD_PSVPAGE: */
            /* case R_PIC30_TBLOFFSET: */
            /* case R_PIC30_PSVOFFSET: */
            /* case R_PIC30_TBLPAGE: */
            /* case R_PIC30_PSVPAGE: */
            /* case R_PIC30_HANDLE: */
            /* case R_PIC30_PADDR: */
            /* case R_PIC30_P_TBLOFFSET: */
            /* case R_PIC30_P_PSVOFFSET: */
            /* case R_PIC30_P_TBLPAGE: */
            /* case R_PIC30_P_PSVPAGE: */
            /* case R_PIC30_P_HANDLE: */
            /* case R_PIC30_P_PADDR: */
            break;
          default:
            {
              *error_msg = (char *) malloc(BUFSIZ);
              sprintf(*error_msg, "Cannot use relocation type %s on a symbol (%s)"
                      " that is located in an executable section.",
                      howto->name, symbol->name);
              return bfd_reloc_dangerous;
            }
          }


      /*
       * if reloc is PC relative, back up the data pointer
       * ...because pic30_coff_perform_pc_relative() advanced it
       */
      if (howto->pc_relative)
        octets -= 4;

      /* Get symbol value.
         Common and undefined weak symbols resolve to zero. */
      if (bfd_is_com_section (symbol->section) ||
          (bfd_is_und_section (symbol->section) && (symbol->flags & BSF_WEAK)))
        relocation = 0;
      else
        relocation = symbol->value;

      reloc_target_output_section = symbol->section->output_section;

      /* Convert input-section-relative symbol value to absolute.  */
      if (output_bfd && howto->partial_inplace == false)
        output_base = 0;
      else
        output_base = reloc_target_output_section->vma;

      relocation += output_base + symbol->section->output_offset;

      /* Add in supplied addend.  */
      relocation += reloc_entry->addend;

#if 0
  printf("symbol->value = %x\n", symbol->value);
  printf("output_base = %x\n", output_base);
  printf("symbol->section->output_offset = %x\n", symbol->section->output_offset);
  printf("reloc_entry->addend = %x\n", reloc_entry->addend);
  printf("relocation = %x\n", relocation);
#endif

      /* Here the variable relocation holds the final address of the
         symbol we are relocating against, plus any addend.  */

      if (howto->pc_relative == true)
        {
          /* This is a PC relative relocation.  We want to set RELOCATION
             to the distance between the address of the symbol and the
             location.  RELOCATION is already the address of the symbol.

             We start by subtracting the address of the section containing
             the location.

             If pcrel_offset is set, we must further subtract the position
             of the location within the section.  Some targets arrange for
             the addend to be the negative of the position of the location
             within the section; for example, i386-aout does this.  For
             i386-aout, pcrel_offset is false.  Some other targets do not
             include the position of the location; for example, m88kbcs,
             or ELF.  For those targets, pcrel_offset is true. */

          relocation -=
            input_section->output_section->vma + input_section->output_offset;

          if (howto->pcrel_offset == true)
            relocation -= (reloc_entry->address / bfd_octets_per_byte (abfd));
        }

      /* FIXME: This overflow checking is incomplete, because the value
         might have overflowed before we get here.  For a correct check we
         need to compute the value in a size larger than bitsize, but we
         can't reasonably do that for a reloc the same size as a host
         machine word.
         FIXME: We should also do overflow checking on the result after
         adding in the value contained in the object file.  */
      if (howto->complain_on_overflow != complain_overflow_dont
          && flag == bfd_reloc_ok)
        flag = bfd_check_overflow (howto->complain_on_overflow,
                                   howto->bitsize,
                                   howto->rightshift,
                                   bfd_arch_bits_per_address (abfd),
                                   relocation);

      /* apply the rightshift parameter */
      relocation >>= (bfd_vma) howto->rightshift;

#if 0
      printf("Final relocation = %lx at %lx\n", relocation, octets);
#endif

      /* Call the pic30-specific range check routine */
      if (howto->complain_on_overflow == complain_overflow_dont
          && flag == bfd_reloc_ok)
        flag = pic30_bfd_reloc_range_check(howto, relocation, error_msg);

      /* Shift everything up to where it's going to be used */
      relocation <<= (bfd_vma) howto->bitpos;

#define DOIT(x) \
  x = ( (x & ~howto->dst_mask) | (((x & howto->src_mask) +  relocation) & howto->dst_mask))

      switch (howto->size)
        {
        case 0:
          {
            char x = bfd_get_8 (abfd, (char *) data + octets);
            DOIT (x);
            bfd_put_8 (abfd, x, (unsigned char *) data + octets);
          }
          break;

        case 1:
          {
            short x = bfd_get_16 (abfd, (bfd_byte *) data + octets);
            DOIT (x);
            bfd_put_16 (abfd, x, (unsigned char *) data + octets);
          }
          break;

        case 2:
          {
            long x = bfd_get_32 (abfd, (bfd_byte *) data + octets);
            DOIT (x);
            bfd_put_32 (abfd, x, (bfd_byte *) data + octets);
          }
          break;

        case 3:
          /* Do nothing */
          break;

        case 4:
#ifdef BFD64
          {
            bfd_vma x = bfd_get_64 (abfd, (bfd_byte *) data + octets);
            DOIT (x);
            bfd_put_64 (abfd, x, (bfd_byte *) data + octets);
          }
#else
          abort ();
#endif
          break;
        default:
          return bfd_reloc_other;
        }

      /*
       * Relocation is complete.  Return the overflow flag to
       * bfd_perform_relocation
       */
      return flag;
    }
}

/*****************************************************************************/

