/* bin2hex - convert an executable object file to HEX format
   using the bfd library. */

#include "bfd.h"
#include <getopt.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "bucomm.h"

#define error -1
#define success 0
#define CHECK_BITS 1
#define EEDATA_BASE_ADDR 0x7ff000

void write_extended_address_record(file_ptr base_address, PTR fp);
void write_section(bfd *abfd, asection *sect, PTR fp);
int write_hex_file(char *name, bfd *abfd);
int dump_symbols(bfd *abfd);
void print_symbol(asymbol *sym);

/* The BFD object file format.  */
static char *target = NULL;

static bfd_size_type actual_prog_used;

const char *version = VERSION;

#if CHECK_BITS
#define MAX_BITS 336
#define TOTAL_BITS 768
#define WORDS_PER_ROW 32
#define OCTETS_PER_WORD 4
#define PHANTOM_BYTE 3

/* list to keep track of partial rows */
struct partial_row
{
  struct partial_row *next;
  bfd_vma addr;
  int bit_count;
  int bit_total;
};
static struct partial_row *partial_row_list;
static int check_bits = 0;
static int total_rows = 0;
static int error_rows = 0;
static unsigned int max_bits = MAX_BITS;
static int ignore_unknown = 0;
static int verbose = 0;

static void pic30_init_list
 PARAMS ((struct partial_row **));

static void pic30_add_partial_row
     PARAMS ((bfd_vma, unsigned int, unsigned int));

static void pic30_free_list
 PARAMS ((struct partial_row **));

static void check_bits_section
 PARAMS ((bfd *, asection *, PTR));

static void check_bits_partial_rows
 PARAMS ((void));
#endif

int
main (argc, argv)
     int argc;
     char **argv;
{
  bfd *abfd;
  char *file;
  char *tool;
  char * dot;
  char * hex_file;
  int c;

  tool = argv[0];
  file = argv[1];

  if (argc == 1)  {
    printf("\n  %s -- convert an object file to Intel hex format\n", argv[0]);
#if CHECK_BITS
    printf ("  %s w/ bit checking", strstr (version, "- ") + 2);
    printf("\n  usage: %s file [options]\n\n", argv[0]);
    printf("  options:\n");
    printf("    -c        check bit density\n");
    printf("    -i        ignore unused words\n");
    printf("    -n MAX    allow MAX bits per row (default = 336)\n");
    printf("    -v        print verbose messages\n");
    printf("\n");
#else
    printf ("  %s", strstr (version, "- ") + 2);
    printf("\n  usage: %s file\n\n", argv[0]);
#endif
    return success;
  }

#if CHECK_BITS
  /* process bit checking options */
  opterr = 0;

  while ((c = getopt (argc, argv, "civn:")) != -1)
    switch (c)
      {
      case 'c':
        check_bits++;
        break;
      case 'i':
        ignore_unknown++;
        break;
      case 'v':
        verbose++;
        break;
      case 'n':
        max_bits = strncmp (optarg, "=", 1) ? atoi(optarg): atoi(optarg+1);
        break;
      case '?':
        if (isprint (optopt))
          fprintf (stderr, "Unknown option `-%c'.\n", optopt);
        else
          fprintf (stderr,
                   "Unknown option character `\\x%x'.\n",
                   optopt);
        return 1;
      default:
        abort ();
      }
#endif

  bfd_init ();

  abfd = bfd_openr (file, target);

  if (abfd == NULL)
    {
      printf("  %s: %s could not be opened.\n", tool, file);
      return error;
    }

  if (bfd_check_format (abfd, bfd_object) == false)
    {
      bfd_close (abfd);
      printf("  %s: %s is not a valid object file.\n", tool, file);
      return error;
    }

  printf("\n");

#if CHECK_BITS
  if (check_bits)
    {
      printf("Checking bit density in %s (max = %d)%s", file, max_bits,
             verbose ? "\n" : " " );

      pic30_init_list(&partial_row_list);

      bfd_map_over_sections(abfd, &check_bits_section, NULL);

      check_bits_partial_rows();

      pic30_free_list(&partial_row_list);

      if (error_rows)
        printf(" %u of %u rows had errors\n", error_rows, total_rows);
      else
        printf(" OK\n");

      bfd_close (abfd);
      return error_rows ? error : success;
    }
#endif

  /* strip extension from filename if present */
  dot = strrchr(file, '.');
  if (dot)
    dot[0] = '\0';

  /* create a new name with .hex extension */
  hex_file = malloc( strlen(file) + strlen(".hex") + 1);
  sprintf( hex_file, "%s%s", file, ".hex");

  if (!write_hex_file(hex_file,abfd)) { /* write it */
    free(hex_file);
    return error;
  }

  printf("\n");

  bfd_close (abfd);
  free(hex_file);
  return success;

}

int
write_hex_file(char *name, bfd *abfd) {
  FILE *fp;
  int rc = 1;

  fp = fopen(name, "w");
  if (!fp) {
    printf("Error: could not open file %s for writing!\n", name);
    return 0;
  }
  
  printf("writing %s\n\n", name);
  printf("section   PC address   byte address   length (w/pad)   actual length  (dec)\n");
  printf("-------   ----------   ------------   --------------   --------------------\n");

  actual_prog_used = 0;

  /* write each section */
  bfd_map_over_sections(abfd, &write_section, fp);

  /* write the end of file marker */
  fprintf(fp, ":00000001FF\n");
  fclose(fp);

  /* write summary info */
  printf("\nTotal program memory used (bytes): %#33lx  (%ld)\n",
         actual_prog_used, actual_prog_used);

  return rc;
}

void
write_section(bfd *abfd, asection *sect, PTR fp) {

  const bfd_size_type row_size=0x10;

  FILE *hexfile;
  unsigned char dat[row_size];
  unsigned char *p;
  file_ptr  PCstart, start, offset;
  file_ptr  upper_addr, base_addr;  /* for Extended Linear Address Records */
  bfd_size_type total, actual;
  unsigned int   num_rows,last_row,i,j,sum;
  unsigned short low_addr;

  base_addr = 0;
  hexfile = fp;
  p = &dat[0];
  offset = 0;
  total = 0;

  /* if section is load-able and has contents */
  if ((sect->flags & SEC_LOAD) && (sect->flags & SEC_HAS_CONTENTS)) {

    /* print section header */
    PCstart = sect->lma;
    start = PCstart << 1;
    total = sect->_raw_size;
    actual = total * .75;

    /* report correct size of EEDATA sections */
    if (PCstart >= EEDATA_BASE_ADDR)
      actual = total / 2;

    printf("%-10s%#10lx%#15lx%#17lx%#16lx  (%ld)\n", sect->name,
           PCstart, start, total, actual, actual);

    actual_prog_used += (total * .75);  /* actual program memory used */
    num_rows = total / row_size;
    last_row = total % row_size;

    /* start with extended address = 0 */
    write_extended_address_record(0,fp);

    /* write the full records */
    for (i=0; i<num_rows; i++) {
      sum = 0;
      low_addr = (unsigned short) (start+offset);
      upper_addr = (start+offset) - low_addr;
      if (upper_addr > base_addr) { /* if address overflows */
        base_addr = upper_addr;
        write_extended_address_record(base_addr,fp);
      }

      bfd_get_section_contents(abfd, sect, p, offset, row_size);
      fprintf(hexfile, ":10%4.4x00", low_addr);  /* begin record */
      sum -= (0x10 + (low_addr>>8) + (low_addr&0xFF));
      offset += row_size;
      for (j=0; j<row_size; j++) {        /* fill in data */
        fprintf(hexfile, "%2.2x", dat[j]);
        sum -= dat[j];
      }
      fprintf(hexfile, "%2.2x\n", (unsigned char) sum);     /* add checksum */
    }

    /* write the last record */
    if (last_row) {
      sum = 0;
      low_addr = (unsigned short) (start+offset);
      upper_addr = (start+offset) - low_addr;
      if (upper_addr > base_addr) { /* if address overflows */
        base_addr = upper_addr;
        write_extended_address_record(base_addr,fp);
      }
      fprintf(hexfile, ":%2.2x%4.4x00", last_row, low_addr);  /* begin record */
      sum -= (last_row + (low_addr>>8) + (low_addr&0xFF));
      bfd_get_section_contents(abfd, sect, p, offset, last_row);
      for (j=0; j<last_row; j++) {  /* fill in data */
        fprintf(hexfile, "%2.2x", dat[j]);
        sum -= dat[j];
      }
      fprintf(hexfile, "%2.2x\n", (unsigned char) sum);        /* add checksum */
    }    
  } /* if ((sect->flags & SEC_LOAD) && (sect->flags & SEC_HAS_CONTENTS)) */
}

void write_extended_address_record(file_ptr base_address, PTR fp) {
  FILE *hexfile;
  unsigned int   sum,upper_addr;

  sum = 0;
  hexfile = fp;
  upper_addr = (unsigned long) base_address>>16;
  sum -= (2 + 0 + 4 + (upper_addr>>8) + (upper_addr&0xFF));
  fprintf(hexfile, ":02000004%4.4x%2.2x\n", upper_addr, (unsigned char) sum);

} /* write_extended_address_record */


int
dump_symbols(bfd *abfd) {
  long size, num, i;
  asymbol **symbols;

  size = bfd_get_symtab_upper_bound(abfd);
  if (size < 0) {
    /* error accessing symbol table */
    return error;
  }

  if (size == 0) {
    printf("  no symbols\n");
    return success;
  }

  symbols = (asymbol **) malloc(size);
  num = bfd_canonicalize_symtab(abfd, symbols);

  if (num < 0) {
    free(symbols);
    /* error processing symbol table */
    return error;
  }
  else {
    for (i=0; i<num; i++) {
      print_symbol(symbols[i]);
    }
    free(symbols);
    return success;
  }
}

void
print_symbol(asymbol *sym) {
  printf("  %s\n", bfd_asymbol_name(sym));
}

#if CHECK_BITS
/*
** Create a new partial_row list
**
** - first item is null
*/
static void
pic30_init_list(lst)
     struct partial_row **lst;
{
  *lst = ((struct partial_row *)
         xmalloc(sizeof(struct partial_row)));
  (*lst)->next = 0;
  (*lst)->addr = 0;
  (*lst)->bit_count = 0;
  (*lst)->bit_total = 0;
}

/*
** Add a partial_row to the list
*/
static void
pic30_add_partial_row (addr, count, total)
     bfd_vma addr;
     unsigned int count, total;
{
  struct partial_row *new, *r, *prev;
  int done = 0;
#if 0
  printf("pic30_add_partial_row: addr = %lx, count = %d, total = %d\n",
         addr, count, total);
#endif
  /*
  ** if this row already exists..
  **   just update the bit count
  */
  prev = partial_row_list;
  for (r = prev; r != NULL; r = r->next)
    {
      if (r && (r->addr == addr))
        {
          r->bit_count += count;
          r->bit_total += total;
          done++;
          break;
        }
      prev = r;
    }

  if (!done)
    {
      /* create a new element */
      new = ((struct partial_row *) 
             xmalloc(sizeof(struct partial_row)));
      new->addr  = addr;
      new->bit_count = count;
      new->bit_total = total;

      /* append it to the list */
      prev->next = new;
      new->next = NULL;
    }
} /* static void pic30_add_partial_row (...) */

/*
** Free a partial_row list
*/
static void
pic30_free_list(lst)
     struct partial_row **lst;
{
  struct partial_row *s, *next;

  if (!(*lst))
    return;

  for (s = *lst; s != NULL; s = next)
    {
      next = s->next;
      free(s);
    }

  *lst = NULL;
} /* static void pic30_free_list (...) */

static void
check_bits_section(abfd, sect, fp)
     bfd *abfd;
     asection *sect;
     PTR fp ATTRIBUTE_UNUSED;
{
  const unsigned int opb = bfd_octets_per_byte (abfd);
  const unsigned int octets_per_row = WORDS_PER_ROW * OCTETS_PER_WORD;
  const unsigned int addr_per_row = octets_per_row / opb;

  unsigned char data[octets_per_row];
  unsigned char *p = &data[0];
  bfd_vma start, offset;
  bfd_size_type octets, units;
  unsigned int num_rows, first_row, last_row;
  unsigned int i,j;

  if (((sect->flags & SEC_LOAD) == 0) ||
      ((sect->flags & SEC_HAS_CONTENTS) == 0))
    return;

  start  = sect->lma;        /* start address           */
  octets = sect->_raw_size;  /* length in octets        */
  units  = octets / opb;     /* length in address units */

  /* compute length (in octets) of 1st partial row */
  if (units < addr_per_row)
    first_row = units * opb;
  else
    {
      j = start % (bfd_vma) addr_per_row;
      first_row = j ? (addr_per_row - j)*opb : 0;
    }

  /* compute number of full rows */
  num_rows = (octets - first_row) / octets_per_row;

  /* compute length (in octets) of last partial row */
  last_row = (octets - first_row) % octets_per_row;

  if (verbose)
    {
      /* print section header */
      j = (first_row ? 1 : 0) + (last_row ? 1 : 0);
      printf("\n  %s: addr = 0x%lx, len = 0x%lx,"
             " full rows = %d, partial rows = %d\n",
             sect->name, start, octets, num_rows, j);
    }
  else
    printf(".");

  /* process first partial row, if any */
  offset = 0;
  if (first_row)
    {
      unsigned int bit_count = 0;
      unsigned int bit_total = 0;

      if (verbose)
        printf("    first row: %8u words\n", first_row / OCTETS_PER_WORD);

      bfd_get_section_contents(abfd, sect, p, offset, first_row);

      /* count the bits in this row */
      for (j = 0; j < first_row; j++)
        {
          int b,d;

          if (((start + offset + j) % OCTETS_PER_WORD) == PHANTOM_BYTE)
            continue;  /* skip phantom bytes */

          bit_total += 8;
          d = data[j];
          for (b = 0; b < 8; b++)
            {
              if (d & 1)
                bit_count++;
              d >>= 1;
            }
        }
      pic30_add_partial_row(start - (start % addr_per_row), bit_count, bit_total);
    }

  /* process whole rows */
  offset += first_row;
  if (num_rows)
    {
      if (verbose)
        printf("    full rows: %8lu words\n",
               (unsigned long) (num_rows * WORDS_PER_ROW));

      for (i = 0; i < num_rows; i++)
        {
          unsigned int bit_count = 0;
          unsigned int bit_total = 0;
          bfd_get_section_contents(abfd, sect, p, offset, octets_per_row);

          /* count the bits in this row */
          for (j = 0; j < octets_per_row; j++)
            {
              int b,d;

              if (((start + offset + j) % OCTETS_PER_WORD) == PHANTOM_BYTE)
                continue;  /* skip phantom bytes */

              bit_total += 8;
              d = data[j];
              for (b = 0; b < 8; b++)
                {
                  if (d & 1)
                    bit_count++;
                  d >>= 1;
                }
            }

          if (bit_count > max_bits)
            {
              error_rows++;
              if (verbose)
                printf("Error: too many bits set (%u) in 0x%lx to 0x%lx\n",
                       bit_count,
                       start + offset,
                       start + offset + octets_per_row);
            }

          total_rows++;
          offset += octets_per_row;
        }
    }

  /* process last partial row, if any */
  if (last_row)
    {
      unsigned int bit_count = 0;
      unsigned int bit_total = 0;

      if (verbose)
        printf("    last row:  %8u words\n", last_row / OCTETS_PER_WORD);

      bfd_get_section_contents(abfd, sect, p, offset, last_row);

      /* count the bits in this row */
      for (j = 0; j < last_row; j++)
        {
          int b,d;

          if (((start + offset + j) % OCTETS_PER_WORD) == PHANTOM_BYTE)
            continue;  /* skip phantom bytes */

          bit_total += 8;
          d = data[j];
          for (b = 0; b < 8; b++)
            {
              if (d & 1)
                bit_count++;
              d >>= 1;
            }
        }
      pic30_add_partial_row(start+offset, bit_count, bit_total);
    }

} /* void check_bits_section (...) */

static void
check_bits_partial_rows()
{
  const unsigned int octets_per_row = WORDS_PER_ROW * OCTETS_PER_WORD;
  struct partial_row *r;

  /*
  ** At this point we've seen all the sections,
  ** and checked all the full rows.
  **
  ** Now check the partial rows.
  */
  if (verbose)
    printf("\n  partial rows (combined):\n");

  if (!partial_row_list)
    return;

  for (r = partial_row_list; r != NULL; r = r->next)
    {
      unsigned int sum;

      if (verbose)
        printf("    0x%lx to 0x%lx, %d/%d\n",
               r->addr, r->addr + octets_per_row, r->bit_count, r->bit_total);

      if (verbose && (r->bit_total < TOTAL_BITS))
        {
          printf("Warning: unknown bits (%d) in 0x%lx to 0x%lx",
                 (TOTAL_BITS - r->bit_total), r->addr, (r->addr + octets_per_row));
          printf("%s", ignore_unknown ? " (ignored)\n" : " (assume 1's)\n");
        }

      /* assume un-programmed octets are 0xFF unless told otherwise */
      sum = r->bit_count + (ignore_unknown ? 0 : (TOTAL_BITS - r->bit_total));

      if (sum > max_bits)
        {
          error_rows++;
          if (verbose)
            printf("Error: too many bits set (%u) in 0x%lx to 0x%lx\n",
                   r->bit_count, r->addr, r->addr + octets_per_row);
        }
      if (r->bit_total)
        total_rows++;
    }

}

#endif

#if 0
/*
   Sample code cloned from coff_object_p in bfd/coffgen.c
   that shows how to read the file header magic number.
   This function can be called after bfd_openr and before
   bfd_check_format. Note that bfd_check_format does way
   more than just check the format. It calls many routines
   that are involved in setting up the bfd.
*/
#include "libbfd.h"
#include <coff/internal.h>
#include "libcoff.h"

unsigned short
coff_file_magic_number (abfd)
     bfd *abfd;
{
  unsigned int filhsz;
  unsigned int aoutsz;
  int nscns;
  PTR filehdr;
  struct internal_filehdr internal_f;
  struct internal_aouthdr internal_a;

  /* figure out how much to read */
  filhsz = bfd_coff_filhsz (abfd);
  aoutsz = bfd_coff_aoutsz (abfd);

  /* get the raw coff header */
  filehdr = bfd_alloc (abfd, filhsz);
  if (filehdr == NULL)
    return 0;
  if (bfd_read (filehdr, 1, filhsz, abfd) != filhsz)
    {
      if (bfd_get_error () != bfd_error_system_call)
	bfd_set_error (bfd_error_wrong_format);
      return 0;
    }
  /* convert to internal form */
  bfd_coff_swap_filehdr_in (abfd, filehdr, &internal_f);
  bfd_release (abfd, filehdr);

  /* read the optional header, too */
  if (internal_f.f_opthdr)
    {
      PTR opthdr;

      opthdr = bfd_alloc (abfd, aoutsz);
      if (opthdr == NULL)
	return 0;;
      if (bfd_read (opthdr, 1, internal_f.f_opthdr, abfd)
	  != internal_f.f_opthdr)
	{
	  return 0;
	}
      bfd_coff_swap_aouthdr_in (abfd, opthdr, (PTR) &internal_a);
      bfd_release (abfd, opthdr);
    }

  return internal_f.f_magic;
}
#endif
